# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import os
import pathlib
import platform
import tempfile
from textwrap import dedent
from unittest.mock import Mock, call, patch

import fixtures
import pytest
from testtools.matchers import DirExists, EndsWith, Equals, Not

from snapcraft.internal.build_providers import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.project import Project

from . import (
    BaseProviderBaseTest,
    MacBaseProviderWithBasesBaseTest,
    ProviderImpl,
    get_project,
)


class BaseProviderTest(BaseProviderBaseTest):
    def test_initialize(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        self.assertThat(provider.project, Equals(self.project))
        self.assertThat(provider.instance_name, Equals(self.instance_name))
        self.assertThat(
            provider.provider_project_dir,
            EndsWith(
                os.path.join("snapcraft", "projects", "project-name", "stub-provider")
            ),
        )
        self.assertThat(
            provider.snap_filename,
            Equals("project-name_1.0_{}.snap".format(self.project.deb_arch)),
        )

    def test_context(self):
        with ProviderImpl(project=self.project, echoer=self.echoer_mock) as provider:
            provider.shell()
            fake_provider = provider

        fake_provider.create_mock.assert_called_once_with("create")
        fake_provider.destroy_mock.assert_called_once_with("destroy")

    def test_context_fails_create(self):
        create_mock = Mock()
        destroy_mock = Mock()

        class BadProviderImpl(ProviderImpl):
            def create(self):
                super().create()
                create_mock("create bad")
                raise errors.ProviderBaseError()

            def destroy(self):
                super().destroy()
                destroy_mock("destroy bad")

        with contextlib.suppress(errors.ProviderBaseError):
            with BadProviderImpl(project=self.project, echoer=self.echoer_mock):
                pass

        create_mock.assert_called_once_with("create bad")
        destroy_mock.assert_called_once_with("destroy bad")

    def test_initialize_snap_filename_with_version(self):
        self.project._snap_meta.version = "test-version"

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        self.assertThat(
            provider.snap_filename,
            Equals("project-name_test-version_{}.snap".format(self.project.deb_arch)),
        )

    def test_launch_instance(self):
        self.useFixture(fixtures.EnvironmentVariable("SNAP_VERSION", "4.0"))

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider.start_mock.side_effect = errors.ProviderInstanceNotFoundError(
            instance_name=self.instance_name
        )
        provider.launch_instance()

        provider.launch_mock.assert_any_call()
        provider.start_mock.assert_any_call()
        provider.save_info_mock.assert_called_once_with(
            {
                "data": {
                    "base": "core16",
                    "created-by-snapcraft-version": "4.0",
                    "host-project-directory": self.project._project_dir,
                }
            }
        )

        self.assertThat(
            provider.run_mock.mock_calls,
            Equals(
                [
                    call(["mv", "/var/tmp/L3Jvb3QvLmJhc2hyYw==", "/root/.bashrc"]),
                    call(["chown", "root:root", "/root/.bashrc"]),
                    call(["chmod", "0600", "/root/.bashrc"]),
                    call(
                        [
                            "mv",
                            "/var/tmp/L2Jpbi9fc25hcGNyYWZ0X3Byb21wdA==",
                            "/bin/_snapcraft_prompt",
                        ]
                    ),
                    call(["chown", "root:root", "/bin/_snapcraft_prompt"]),
                    call(["chmod", "0755", "/bin/_snapcraft_prompt"]),
                    call(
                        [
                            "mv",
                            "/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0",
                            "/etc/apt/sources.list",
                        ]
                    ),
                    call(["chown", "root:root", "/etc/apt/sources.list"]),
                    call(["chmod", "0644", "/etc/apt/sources.list"]),
                    call(
                        [
                            "mv",
                            "/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0LmQvZGVmYXVsdC5zb3VyY2Vz",
                            "/etc/apt/sources.list.d/default.sources",
                        ]
                    ),
                    call(
                        [
                            "chown",
                            "root:root",
                            "/etc/apt/sources.list.d/default.sources",
                        ]
                    ),
                    call(["chmod", "0644", "/etc/apt/sources.list.d/default.sources"]),
                    call(
                        [
                            "mv",
                            "/var/tmp/L2V0Yy9hcHQvc291cmNlcy5saXN0LmQvZGVmYXVsdC1zZWN1cml0eS5zb3VyY2Vz",
                            "/etc/apt/sources.list.d/default-security.sources",
                        ]
                    ),
                    call(
                        [
                            "chown",
                            "root:root",
                            "/etc/apt/sources.list.d/default-security.sources",
                        ]
                    ),
                    call(
                        [
                            "chmod",
                            "0644",
                            "/etc/apt/sources.list.d/default-security.sources",
                        ]
                    ),
                    call(
                        [
                            "mv",
                            "/var/tmp/L2V0Yy9hcHQvYXB0LmNvbmYuZC8wMC1zbmFwY3JhZnQ=",
                            "/etc/apt/apt.conf.d/00-snapcraft",
                        ]
                    ),
                    call(["chown", "root:root", "/etc/apt/apt.conf.d/00-snapcraft"]),
                    call(["chmod", "0644", "/etc/apt/apt.conf.d/00-snapcraft"]),
                    call(["apt-get", "update"]),
                    call(["apt-get", "dist-upgrade", "--yes"]),
                    call(["apt-get", "install", "--yes", "apt-transport-https"]),
                    call(["snap", "unset", "system", "proxy.http"]),
                    call(["snap", "unset", "system", "proxy.https"]),
                ]
            ),
        )

        self.assertThat(provider.provider_project_dir, DirExists())

    def test_launch_instance_with_proxies(self):
        provider = ProviderImpl(
            project=self.project,
            echoer=self.echoer_mock,
            build_provider_flags={
                "http_proxy": "http://1.2.3.4:8080",
                "https_proxy": "http://2.3.4.5:8080",
            },
        )
        provider.launch_instance()

        provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", "proxy.http=http://1.2.3.4:8080"]),
                call(["snap", "set", "system", "proxy.https=http://2.3.4.5:8080"]),
            ]
        )

    def test_launch_instance_with_cert_file(self):
        test_certs = pathlib.Path(self.path, "certs")
        test_certs.mkdir()
        test_cert1 = pathlib.Path(test_certs, "1.crt")
        test_cert1.write_text("")

        provider = ProviderImpl(
            project=self.project,
            echoer=self.echoer_mock,
            build_provider_flags={"SNAPCRAFT_ADD_CA_CERTIFICATES": str(test_cert1)},
        )
        provider.launch_instance()

        provider.push_file_mock.assert_has_calls(
            [
                call(
                    destination="/usr/local/share/ca-certificates/1.crt",
                    source=str(test_cert1),
                )
            ]
        )
        provider.run_mock.assert_has_calls([call(["update-ca-certificates"])])

    def test_launch_instance_with_cert_dir_files(self):
        test_certs = pathlib.Path(self.path, "certs")
        test_certs.mkdir()
        test_cert1 = pathlib.Path(test_certs, "1.crt")
        test_cert1.write_text("")
        test_cert2 = pathlib.Path(test_certs, "2.crt")
        test_cert2.write_text("")

        provider = ProviderImpl(
            project=self.project,
            echoer=self.echoer_mock,
            build_provider_flags={"SNAPCRAFT_ADD_CA_CERTIFICATES": str(test_certs)},
        )
        provider.launch_instance()

        provider.push_file_mock.assert_has_calls(
            [
                call(
                    destination="/usr/local/share/ca-certificates/1.crt",
                    source=str(test_cert1),
                ),
                call(
                    destination="/usr/local/share/ca-certificates/2.crt",
                    source=str(test_cert2),
                ),
            ]
        )
        provider.run_mock.assert_has_calls([call(["update-ca-certificates"])])

    def test_expose_prime(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider.expose_prime()

        provider.mount_mock.assert_called_once_with(
            self.project.prime_dir, "/root/prime"
        )

    @patch("pathlib.Path.home", return_value=pathlib.Path("/home/user"))
    def test_bind_ssh(self, mock_home):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        # False.
        provider.build_provider_flags = dict(SNAPCRAFT_BIND_SSH=False)
        provider._mount_project()
        provider.mount_mock.assert_has_calls(
            [call(self.project._project_dir, "/root/project")]
        )

        # Not present.
        provider.build_provider_flags = dict()
        provider._mount_project()
        provider.mount_mock.assert_has_calls(
            [call(self.project._project_dir, "/root/project")]
        )

        # True.
        provider.build_provider_flags = dict(SNAPCRAFT_BIND_SSH=True)
        provider._mount_project()
        provider.mount_mock.assert_has_calls(
            [
                call(self.project._project_dir, "/root/project"),
                call("/home/user/.ssh", "/root/.ssh"),
            ]
        )

    def test_start_instance(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        provider.launch_instance()

        provider.launch_mock.assert_not_called()
        provider.start_mock.assert_any_call()

        # Verify the updates that take place on each start.
        provider.run_mock.mock_calls == [
            call(["snap", "unset", "system", "proxy.http"]),
            call(["snap", "unset", "system", "proxy.https"]),
        ]

        # Given the way we constructe this test, this directory should not exist
        # TODO add robustness to start. (LP: #1792242)
        self.assertThat(provider.provider_project_dir, Not(DirExists()))

    def test_clean_part(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        provider.clean_parts(part_names=("part1",))

        provider.run_mock.assert_called_once_with(["snapcraft", "clean", "part1"])

    def test_clean_multiple_parts(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)

        provider.clean_parts(part_names=("part1", "part2"))

        provider.run_mock.assert_called_once_with(
            ["snapcraft", "clean", "part1", "part2"]
        )

    def test_passthrough_environment_flags_empty(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider.build_provider_flags = dict()

        results = provider._get_env_command()

        self.assertThat(
            results,
            Equals(
                [
                    "env",
                    "SNAPCRAFT_BUILD_ENVIRONMENT=managed-host",
                    "HOME=/root",
                    "SNAPCRAFT_HAS_TTY=False",
                ]
            ),
        )

    def test_passthrough_environment_flags_non_string(self):
        # Set http_proxy to a bool even though it doesn't make sense...
        # This is to verify non-strings are treated as strings OK for environ.
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider.build_provider_flags = dict(http_proxy=True)

        results = provider._get_env_command()

        self.assertThat(
            results,
            Equals(
                [
                    "env",
                    "SNAPCRAFT_BUILD_ENVIRONMENT=managed-host",
                    "HOME=/root",
                    "SNAPCRAFT_HAS_TTY=False",
                    "http_proxy=True",
                ]
            ),
        )

    def test_passthrough_environment_flags_all(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider.build_provider_flags = dict(
            http_proxy="http://127.0.0.1:8080",
            https_proxy="http://127.0.0.1:8080",
            SNAPCRAFT_BUILD_INFO=True,
            SNAPCRAFT_IMAGE_INFO='{"build_url":"https://example.com"}',
        )

        results = provider._get_env_command()

        self.assertThat(
            results,
            Equals(
                [
                    "env",
                    "SNAPCRAFT_BUILD_ENVIRONMENT=managed-host",
                    "HOME=/root",
                    "SNAPCRAFT_HAS_TTY=False",
                    "http_proxy=http://127.0.0.1:8080",
                    "https_proxy=http://127.0.0.1:8080",
                    "SNAPCRAFT_BUILD_INFO=True",
                    'SNAPCRAFT_IMAGE_INFO={"build_url":"https://example.com"}',
                ]
            ),
        )


class BaseProviderProvisionSnapcraftTest(BaseProviderBaseTest):
    def test_setup_snapcraft(self):
        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider._setup_snapcraft()

        self.snap_injector_mock.assert_called_once_with(
            registry_filepath=os.path.join(
                provider.provider_project_dir, "snap-registry.yaml"
            ),
            runner=provider._run,
            file_pusher=provider._push_file,
            inject_from_host=True,
        )
        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="snapd"),
                call(snap_name="core16"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(4))
        self.snap_injector_mock().apply.assert_called_once_with()

    def test_ephemeral_setup_snapcraft(self):
        provider = ProviderImpl(
            project=self.project, echoer=self.echoer_mock, is_ephemeral=True
        )
        provider._setup_snapcraft()

        self.snap_injector_mock.assert_called_once_with(
            registry_filepath=os.path.join(
                provider.provider_project_dir, "snap-registry.yaml"
            ),
            runner=provider._run,
            file_pusher=provider._push_file,
            inject_from_host=True,
        )
        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="snapd"),
                call(snap_name="core16"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(4))
        self.snap_injector_mock().apply.assert_called_once_with()

    def test_setup_snapcraft_with_core(self):
        self.project._snap_meta.base = "core"
        self.project._snap_meta.confinement = "classic"

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider._setup_snapcraft()

        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="core"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(3))
        self.snap_injector_mock().apply.assert_called_once_with()

    def test_setup_snapcraft_for_classic_build(self):
        self.project._snap_meta.base = "core18"
        self.project._snap_meta.confinement = "classic"

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider._setup_snapcraft()

        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="snapd"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(3))
        self.snap_injector_mock().apply.assert_called_once_with()


class MacProviderProvisionSnapcraftTest(MacBaseProviderWithBasesBaseTest):
    def test_setup_snapcraft_with_base(self):
        self.project._snap_meta.base = "core18"

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider._setup_snapcraft()

        self.snap_injector_mock.assert_called_once_with(
            registry_filepath=os.path.join(
                provider.provider_project_dir, "snap-registry.yaml"
            ),
            runner=provider._run,
            file_pusher=provider._push_file,
            inject_from_host=False,
        )
        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="snapd"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(3))
        self.snap_injector_mock().apply.assert_called_once_with()

    def test_setup_snapcraft_for_classic_build(self):
        self.project._snap_meta.base = "core18"
        self.project._snap_meta.confinement = "classic"

        provider = ProviderImpl(project=self.project, echoer=self.echoer_mock)
        provider._setup_snapcraft()

        self.snap_injector_mock.assert_called_once_with(
            registry_filepath=os.path.join(
                provider.provider_project_dir, "snap-registry.yaml"
            ),
            runner=provider._run,
            file_pusher=provider._push_file,
            inject_from_host=False,
        )
        self.snap_injector_mock().add.assert_has_calls(
            [
                call(snap_name="snapd"),
                call(snap_name="core18"),
                call(snap_name="snapcraft"),
            ]
        )
        self.assertThat(self.snap_injector_mock().add.call_count, Equals(3))
        self.snap_injector_mock().apply.assert_called_once_with()


class TestCompatibilityClean:
    scenarios = [
        (
            "same-base-no-clean",
            dict(
                base="core16",
                loaded_info={"base": "core16", "created-by-snapcraft-version": "1.0"},
                version="1.0",
                expect_clean=False,
            ),
        ),
        (
            "different-base-clean",
            dict(
                base="core16",
                loaded_info={"base": "core18", "created-by-snapcraft-version": "1.0"},
                version="1.0",
                expect_clean=True,
            ),
        ),
        (
            "same-project-dir-no-clean",
            dict(
                base="core18",
                loaded_info={
                    "base": "core18",
                    "created-by-snapcraft-version": "1.0",
                    "host-project-directory": "/fake/host/dir",
                },
                version="1.0",
                expect_clean=False,
            ),
        ),
        (
            "no-project-dir-no-clean",
            dict(
                base="core18",
                loaded_info={"base": "core18", "created-by-snapcraft-version": "1.0"},
                version="1.0",
                expect_clean=False,
            ),
        ),
        (
            "different-project-dir-clean",
            dict(
                base="core18",
                loaded_info={
                    "base": "core18",
                    "created-by-snapcraft-version": "1.0",
                    "host-project-directory": "/nowhere",
                },
                version="1.0",
                expect_clean=True,
            ),
        ),
        (
            "unspecified-base-clean",
            dict(
                base="core20",
                loaded_info={"created-by-snapcraft-version": "1.0"},
                version="1.0",
                expect_clean=True,
            ),
        ),
        (
            "unspecified-created-version-clean",
            dict(
                base="core20",
                loaded_info={"base": "core20"},
                version="1.0",
                expect_clean=True,
            ),
        ),
        (
            "downgrade-version-clean",
            dict(
                base="core20",
                loaded_info={"base": "core20", "created-by-snapcraft-version": "2.0"},
                version="1.0",
                expect_clean=True,
            ),
        ),
        (
            "same-version-no-clean",
            dict(
                base="core20",
                loaded_info={"base": "core20", "created-by-snapcraft-version": "2.0"},
                version="2.0",
                expect_clean=False,
            ),
        ),
        (
            "upgrade-version-no-clean",
            dict(
                base="core20",
                loaded_info={"base": "core20", "created-by-snapcraft-version": "2.0"},
                version="3.0",
                expect_clean=False,
            ),
        ),
    ]

    def test_scenario(
        self, monkeypatch, in_snap, base, loaded_info, version, expect_clean
    ):
        monkeypatch.setattr("os.getcwd", lambda: "/fake/host/dir")
        monkeypatch.setenv("SNAP_VERSION", version)

        provider = ProviderImpl(project=get_project(), echoer=Mock())
        provider.project._snap_meta.base = base
        provider.loaded_info = loaded_info

        provider._ensure_compatible_build_environment()

        if expect_clean:
            provider.clean_project_mock.assert_called_once_with()
        else:
            provider.clean_project_mock.assert_not_called()


_ARCHIVES = {
    "x86": {
        "main": "http://archive.ubuntu.com/ubuntu",
        "security": "http://security.ubuntu.com/ubuntu",
    },
    "ports": {
        "main": "http://ports.ubuntu.com/ubuntu-ports",
        "security": "http://ports.ubuntu.com/ubuntu-ports",
    },
}


@pytest.mark.parametrize(
    "machine_platform",
    [
        ("x86_64", _ARCHIVES["x86"]),
        ("AMD64", _ARCHIVES["x86"]),
        ("aarch64", _ARCHIVES["ports"]),
    ],
)
@pytest.mark.parametrize(
    "distro", [("core", "xenial"), ("core18", "bionic"), ("core20", "focal")]
)
def test_setup_environment_content_x86(
    tmp_work_path, monkeypatch, machine_platform, distro
):
    snapcraft_project = Project()
    snapcraft_project._snap_meta = Snap(name="test-snap", base=distro[0])

    monkeypatch.setattr(platform, "machine", lambda: machine_platform[0])

    recorded_files = dict()

    @contextlib.contextmanager
    def fake_namedtempfile(*, suffix: str, **kwargs):
        # Usage hides the file basename in the suffix.
        tmp_path = os.path.join("tmpfile")
        with open(tmp_path, "wb") as f_write:
            yield f_write
        with open(tmp_path, "r") as f_read:
            recorded_files[suffix] = f_read.read()

    monkeypatch.setattr(tempfile, "NamedTemporaryFile", fake_namedtempfile)

    provider = ProviderImpl(project=snapcraft_project, echoer=Mock())
    provider._setup_environment()

    assert recorded_files == {
        ".bashrc": '#!/bin/bash\nexport PS1="\\h \\$(/bin/_snapcraft_prompt)# "\n',
        "00-snapcraft": 'Apt::Install-Recommends "false";\n',
        "_snapcraft_prompt": dedent(
            """\
            #!/bin/bash
            if [[ "$PWD" =~ ^$HOME.* ]]; then
                path="${PWD/#$HOME/\\ ..}"
                if [[ "$path" == " .." ]]; then
                    ps1=""
                else
                    ps1="$path"
                fi
            else
                ps1="$PWD"
            fi
            echo -n $ps1
            """
        ),
        "default.sources": dedent(
            f"""\
                Types: deb
                URIs: {machine_platform[1]["main"]}
                Suites: {distro[1]} {distro[1]}-updates
                Components: main multiverse restricted universe
            """
        ),
        "default-security.sources": dedent(
            f"""\
                Types: deb
                URIs: {machine_platform[1]["security"]}
                Suites: {distro[1]}-security
                Components: main multiverse restricted universe
            """
        ),
        "sources.list": "",
    }



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.storeapi import channels


class TestChannel:
    scenarios = (
        (
            "risk",
            dict(
                channel="edge",
                expected_track="latest",
                expected_risk="edge",
                expected_branch=None,
                expected_channel="edge",
            ),
        ),
        (
            "track/risk",
            dict(
                channel="11/candidate",
                expected_track="11",
                expected_risk="candidate",
                expected_branch=None,
                expected_channel="11/candidate",
            ),
        ),
        (
            "track/risk/branch",
            dict(
                channel="11/edge/patch-1",
                expected_track="11",
                expected_risk="edge",
                expected_branch="patch-1",
                expected_channel="11/edge/patch-1",
            ),
        ),
        (
            "risk/branch",
            dict(
                channel="edge/patch-2",
                expected_track="latest",
                expected_risk="edge",
                expected_branch="patch-2",
                expected_channel="edge/patch-2",
            ),
        ),
    )

    def test_channel(
        self, channel, expected_track, expected_risk, expected_branch, expected_channel
    ):
        c = channels.Channel(channel)

        assert c.track == expected_track
        assert c.risk == expected_risk
        assert c.branch == expected_branch
        assert str(c) == expected_channel
        assert repr(c), "'{}'".format(expected_channel)


class TestChannelTuple:
    scenarios = (
        (
            "risk",
            dict(
                track=None,
                risk="edge",
                branch=None,
                expected_track="latest",
                expected_channel="edge",
            ),
        ),
        (
            "track/risk",
            dict(
                track="11",
                risk="candidate",
                branch=None,
                expected_track="11",
                expected_channel="11/candidate",
            ),
        ),
        (
            "track/risk/branch",
            dict(
                track="11",
                risk="edge",
                branch="patch-1",
                expected_track="11",
                expected_channel="11/edge/patch-1",
            ),
        ),
        (
            "risk/branch",
            dict(
                track="latest",
                risk="edge",
                branch="patch-2",
                expected_track="latest",
                expected_channel="latest/edge/patch-2",
            ),
        ),
    )

    def test_channel(self, track, risk, branch, expected_track, expected_channel):
        c = channels.Channel.from_channel_tuple(track=track, risk=risk, branch=branch)
        assert c.track == expected_track
        assert c.risk == risk
        assert c.branch == branch
        assert str(c) == expected_channel
        assert repr(c), "'{}'".format(expected_channel)


class TestInvalidChannel:
    scenarios = (
        ("invalid risk", dict(channel="invalid-edge")),
        ("invalid track/risk", dict(channel="track/invalid-edge")),
        ("invalid track/risk/branch", dict(channel="track/invalid-edge/branch")),
        ("invalid tuple", dict(channel="start/track/invalid-edge/branch/end")),
    )

    def test_invalid(self, channel):
        with pytest.raises(RuntimeError):
            channels.Channel(channel)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import fixtures
from testtools.matchers import Contains, Equals

import snapcraft
from tests import fixture_setup

from . import CommandBaseTestCase


class ListPluginsCommandTestCase(CommandBaseTestCase):

    command_name = "list-plugins"

    default_plugin_output = (
        "fake      dotruby         kernelz  checkbox-provider  oxide  hornet      dumper  "
        "jamesbuild  none   macaroon    gmake\nnot-real  godependencies  planton  "
        "snake              web    fake-tools  cradle  maker       pmake  breaktools\n"
    )

    # fake plugin list, made up names so it is obvious when it fails that
    # we are not using real data, but use names to have some sort of simulation
    # with reality
    plugin_list = [
        "fake",
        "not-real",
        "dotruby",
        "godependencies",
        "kernelz",
        "planton",
        "checkbox-provider",
        "snake",
        "oxide",
        "web",
        "hornet",
        "fake-tools",
        "dumper",
        "cradle",
        "jamesbuild",
        "maker",
        "none",
        "pmake",
        "macaroon",
        "breaktools",
        "gmake",
    ]

    def setUp(self):
        super().setUp()

        self.maxDiff = None

        # Generate fake plugin data
        iter_modules_result = (("", p, "") for p in self.plugin_list)

        self.fake_iter_modules = fixtures.MockPatch(
            "pkgutil.iter_modules", return_value=iter_modules_result
        )
        self.useFixture(self.fake_iter_modules)

    def test_default_from_snapcraft_yaml(self):
        self.useFixture(
            fixture_setup.SnapcraftYaml(
                self.path,
                base="core18",
                parts={"part1": {"source": ".", "plugin": "nil"}},
            )
        )

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Displaying plugins available for 'core18")
        )
        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v1.__path__
        )

    def test_alias(self):
        self.command_name = "plugins"
        self.useFixture(
            fixture_setup.SnapcraftYaml(
                self.path,
                base="core18",
                parts={"part1": {"source": ".", "plugin": "nil"}},
            )
        )

        result = self.run_command([self.command_name])

        self.assertThat(result.exit_code, Equals(0))

    def test_core20_list(self):
        result = self.run_command([self.command_name, "--base", "core20"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Contains("Displaying plugins available for 'core20")
        )

        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v2.__path__
        )

    def test_core2y_list(self):
        # Note that core2y is some future base, _not_ allowed to be used from cmdline
        # This tests that addition of the next base will use the latest version of plugins
        snapcraft.cli.discovery.list_plugins.callback("core2y")
        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v2.__path__
        )

    def test_list_plugins_non_tty(self):
        fake_terminal = fixture_setup.FakeTerminal(isatty=False)
        self.useFixture(fake_terminal)

        result = self.run_command([self.command_name, "--base", "core18"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(self.default_plugin_output))
        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v1.__path__
        )

    def test_list_plugins_large_terminal(self):
        fake_terminal = fixture_setup.FakeTerminal(columns=999)
        self.useFixture(fake_terminal)

        result = self.run_command([self.command_name, "--base", "core18"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(self.default_plugin_output))
        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v1.__path__
        )

    def test_list_plugins_small_terminal(self):
        fake_terminal = fixture_setup.FakeTerminal(columns=60)
        self.useFixture(fake_terminal)

        expected_output = [
            "fake            checkbox-provider  dumper      macaroon",
            "not-real        snake              cradle      breaktools",
            "dotruby         oxide              jamesbuild  gmake",
            "godependencies  web                maker",
            "kernelz         hornet             none",
            "planton         fake-tools         pmake",
        ]

        result = self.run_command([self.command_name, "--base", "core18"])

        self.assertThat(result.exit_code, Equals(0))
        output_slice = [o.strip() for o in result.output.splitlines()][1:]
        self.assertThat(output_slice, Equals(expected_output))
        self.fake_iter_modules.mock.assert_called_once_with(
            snapcraft.plugins.v1.__path__
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import io
import json
from textwrap import dedent
from unittest.mock import Mock, call, patch

import pytest
from macaroonbakery import bakery, httpbakery
from pymacaroons.macaroon import Macaroon

from snapcraft.storeapi.http_clients._candid_client import (
    CandidClient,
    CandidConfig,
    _http_client,
)


def test_config_section_name():
    assert CandidConfig()._get_section_name() == "dashboard.snapcraft.io"


def test_config_section_name_with_env(monkeypatch):
    monkeypatch.setenv("STORE_DASHBOARD_URL", "http://dashboard.other.com")

    assert CandidConfig()._get_section_name() == "dashboard.other.com"


def test_config_path(xdg_dirs):
    assert (
        CandidConfig()._get_config_path() == xdg_dirs / ".config/snapcraft/candid.cfg"
    )


def test_candid_client_has_no_credentials(xdg_dirs):
    assert CandidClient().has_credentials() is False


def test_candid_client_has_credentials(xdg_dirs):
    # Baseline check.
    assert CandidClient().has_credentials() is False

    # Setup.
    client = CandidClient()
    client._macaroon = "macaroon"
    client._auth = "auth"

    assert client.has_credentials() is True
    assert CandidClient().has_credentials() is True


@pytest.fixture
def candid_client(xdg_dirs, monkeypatch):
    """Return a CandidClient with an alterate requests method."""
    bakery_client = Mock(spec=httpbakery.Client)

    def mock_discharge(*args, **kwargs):
        return [
            Macaroon(
                location="api.snapcraft.io",
                signature="d9533461d7835e4851c7e3b639144406cf768597dea6e133232fbd2385a5c050",
            )
        ]

    monkeypatch.setattr(bakery, "discharge_all", mock_discharge)

    return CandidClient(bakery_client=bakery_client)


@pytest.fixture
def snapcraft_macaroon():
    return json.dumps(
        {
            "s64": "a0Vi7CwhHWjS4bxzKPhCZQIEJDvlbh9FyhOtWx0tNFQ",
            "c": [
                {"i": "time-before 2022-03-18T19:54:57.151721Z"},
                {
                    "v64": "pDqaL9KDrPfCQCLDUdPc8yO2bTQheWGsM1tpxRaS_4BT3r6zpdnT5TelXz8vpjb4iUhTnc60-x5DPKJOpRuwAi4qMdNa67Vo",
                    "l": "https://api.jujucharms.com/identity/",
                    "i64": "AoZh2j7mbDQgh3oK3qMqoXKKFAnJvmOKwmDCNYHIxHqQnFLJZJUBpqoiJtqra-tyXPPMUTmfuXMgOWP7xKwTD26FBgtJBdh1mE1wt3kf0Ur_TnOzbAWQCHKxqK9jAp1jYv-LlLLAlQAmoqvz9fBf2--dIxHiLIRTThmAESAnlLZHOJ7praDmIScsLQC475a85avA",
                },
                {
                    "i": 'extra {"package_id": null, "channel": null, "acl": ["package_access", "package_manage", "package_push", "package_register", "package_release", "package_update"], "store_ids": null}'
                },
            ],
            "l": "api.snapcraft.io",
            "i64": "AwoQ2Ft5YBjnovqdr8VNV3TSlhIBMBoOCgVsb2dpbhIFbG9naW4",
        }
    )


def test_login_discharge_macaroon(candid_client, snapcraft_macaroon):
    candid_client.request
    candid_client.login(macaroon=snapcraft_macaroon)

    assert candid_client.has_credentials() is True
    assert candid_client._macaroon == snapcraft_macaroon
    assert candid_client._auth == (
        "W3siaWRlbnRpZmllciI6ICIiLCAic2lnbmF0dXJlIjogImQ5NTMzNDYxZDc4MzVlNDg1MWM"
        "3ZTNiNjM5MTQ0NDA2Y2Y3Njg1OTdkZWE2ZTEzMzIzMmZiZDIzODVhNWMwNTAiLCAibG9jYX"
        "Rpb24iOiAiYXBpLnNuYXBjcmFmdC5pbyJ9XQ=="
    )


def test_login_discharge_macaroon_no_save(candid_client, snapcraft_macaroon):
    candid_client.login(macaroon=snapcraft_macaroon, save=False)

    assert candid_client.has_credentials() is False
    assert candid_client._macaroon == snapcraft_macaroon
    assert candid_client._auth == (
        "W3siaWRlbnRpZmllciI6ICIiLCAic2lnbmF0dXJlIjogImQ5NTMzNDYxZDc4MzVlNDg1MWM"
        "3ZTNiNjM5MTQ0NDA2Y2Y3Njg1OTdkZWE2ZTEzMzIzMmZiZDIzODVhNWMwNTAiLCAibG9jYX"
        "Rpb24iOiAiYXBpLnNuYXBjcmFmdC5pbyJ9XQ=="
    )


def test_login_with_config_fd(candid_client, snapcraft_macaroon):
    with io.StringIO() as config_fd:
        print("[dashboard.snapcraft.io]", file=config_fd)
        print(f"macaroon = {snapcraft_macaroon}", file=config_fd)
        print(f"auth = 1234567890noshare", file=config_fd)
        config_fd.seek(0)

        candid_client.login(config_fd=config_fd)

    assert candid_client.has_credentials() is True
    assert candid_client._macaroon == snapcraft_macaroon
    assert candid_client._auth == "1234567890noshare"


def test_login_with_config_fd_no_save(candid_client, snapcraft_macaroon):
    with io.StringIO() as config_fd:
        print("[dashboard.snapcraft.io]", file=config_fd)
        print(f"macaroon = {snapcraft_macaroon}", file=config_fd)
        print(f"auth = 1234567890noshare", file=config_fd)
        config_fd.seek(0)

        candid_client.login(config_fd=config_fd, save=False)

    assert candid_client.has_credentials() is False
    assert candid_client._macaroon == snapcraft_macaroon
    assert candid_client._auth == "1234567890noshare"


@pytest.fixture
def authed_client(candid_client, snapcraft_macaroon):
    candid_client.login(macaroon=snapcraft_macaroon)
    assert candid_client.has_credentials() is True

    return candid_client


def test_logout(authed_client):
    authed_client.logout()

    assert authed_client.has_credentials() is False


def test_export_login(authed_client):
    with io.StringIO() as config_fd:
        authed_client.export_login(config_fd=config_fd, encode=False)

        config_fd.seek(0)

        assert config_fd.getvalue().strip() == dedent(
            f"""\
        [dashboard.snapcraft.io]
        auth = {authed_client._auth}
        macaroon = {authed_client._macaroon}"""
        )


def test_export_login_base64_encoded(authed_client):
    with io.StringIO() as config_fd:
        authed_client.export_login(config_fd=config_fd, encode=True)

        config_fd.seek(0)

        assert config_fd.getvalue().strip() == (
            "W2Rhc2hib2FyZC5zbmFwY3JhZnQuaW9dCmF1dGggPSBXM3NpYVdSbGJuUnBabWxsY2lJNklDSWlMQ0FpY"
            "zJsbmJtRjBkWEpsSWpvZ0ltUTVOVE16TkRZeFpEYzRNelZsTkRnMU1XTTNaVE5pTmpNNU1UUTBOREEyWT"
            "JZM05qZzFPVGRrWldFMlpURXpNekl6TW1aaVpESXpPRFZoTldNd05UQWlMQ0FpYkc5allYUnBiMjRpT2l"
            "BaVlYQnBMbk51WVhCamNtRm1kQzVwYnlKOVhRPT0KbWFjYXJvb24gPSB7InM2NCI6ICJhMFZpN0N3aEhX"
            "alM0Ynh6S1BoQ1pRSUVKRHZsYmg5RnloT3RXeDB0TkZRIiwgImMiOiBbeyJpIjogInRpbWUtYmVmb3JlI"
            "DIwMjItMDMtMThUMTk6NTQ6NTcuMTUxNzIxWiJ9LCB7InY2NCI6ICJwRHFhTDlLRHJQZkNRQ0xEVWRQYz"
            "h5TzJiVFFoZVdHc00xdHB4UmFTXzRCVDNyNnpwZG5UNVRlbFh6OHZwamI0aVVoVG5jNjAteDVEUEtKT3B"
            "SdXdBaTRxTWROYTY3Vm8iLCAibCI6ICJodHRwczovL2FwaS5qdWp1Y2hhcm1zLmNvbS9pZGVudGl0eS8i"
            "LCAiaTY0IjogIkFvWmgyajdtYkRRZ2gzb0szcU1xb1hLS0ZBbkp2bU9Ld21EQ05ZSEl4SHFRbkZMSlpKV"
            "UJwcW9pSnRxcmEtdHlYUFBNVVRtZnVYTWdPV1A3eEt3VEQyNkZCZ3RKQmRoMW1FMXd0M2tmMFVyX1RuT3"
            "piQVdRQ0hLeHFLOWpBcDFqWXYtTGxMTEFsUUFtb3F2ejlmQmYyLS1kSXhIaUxJUlRUaG1BRVNBbmxMWkh"
            "PSjdwcmFEbUlTY3NMUUM0NzVhODVhdkEifSwgeyJpIjogImV4dHJhIHtcInBhY2thZ2VfaWRcIjogbnVs"
            "bCwgXCJjaGFubmVsXCI6IG51bGwsIFwiYWNsXCI6IFtcInBhY2thZ2VfYWNjZXNzXCIsIFwicGFja2FnZ"
            "V9tYW5hZ2VcIiwgXCJwYWNrYWdlX3B1c2hcIiwgXCJwYWNrYWdlX3JlZ2lzdGVyXCIsIFwicGFja2FnZV"
            "9yZWxlYXNlXCIsIFwicGFja2FnZV91cGRhdGVcIl0sIFwic3RvcmVfaWRzXCI6IG51bGx9In1dLCAibCI"
            "6ICJhcGkuc25hcGNyYWZ0LmlvIiwgImk2NCI6ICJBd29RMkZ0NVlCam5vdnFkcjhWTlYzVFNsaElCTUJv"
            "T0NnVnNiMmRwYmhJRmJHOW5hVzQifQoK"
        )


@pytest.fixture
def request_mock():
    patched = patch.object(
        _http_client.Client, "request", spec=_http_client.Client.request
    )
    try:
        yield patched.start()
    finally:
        patched.stop()


@pytest.mark.parametrize("method", ["GET", "PUT", "POST"])
@pytest.mark.parametrize("params", [None, {}, {"foo": "bar"}])
def test_request(authed_client, request_mock, method, params):
    authed_client.request(method, "https://dashboard.snapcraft.io/foo", params=params)

    assert request_mock.mock_calls == [
        call(
            method,
            "https://dashboard.snapcraft.io/foo",
            params=params,
            headers={"Macaroons": authed_client._auth},
        ),
        call().ok.__bool__(),
    ]


def test_request_with_headers(authed_client, request_mock):
    authed_client.request(
        "GET", "https://dashboard.snapcraft.io/foo", headers={"foo": "bar"}
    )

    assert request_mock.mock_calls == [
        call(
            "GET",
            "https://dashboard.snapcraft.io/foo",
            params=None,
            headers={"foo": "bar", "Macaroons": authed_client._auth},
        ),
        call().ok.__bool__(),
    ]


@pytest.mark.parametrize("method", ["GET", "PUT", "POST"])
@pytest.mark.parametrize("params", [None, {}, {"foo": "bar"}])
@pytest.mark.parametrize("headers", [None, {}, {"foo": "bar"}])
def test_request_no_auth(authed_client, request_mock, method, params, headers):
    authed_client.request(
        method,
        "https://dashboard.snapcraft.io/foo",
        params=params,
        headers=headers,
        auth_header=False,
    )

    assert request_mock.mock_calls == [
        call(
            method, "https://dashboard.snapcraft.io/foo", params=params, headers=headers
        ),
        call().ok.__bool__(),
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from typing import Any, Dict, Optional, Tuple

from .. import errors

_PLATFORM_SNAP = dict(core18="gnome-3-28-1804")
_DOC_TEMPLATE = textwrap.dedent(
    """\
   This extension eases creation of snaps that use Flutter

    It tracks the {channel} channel for flutter.

    At build time it ensures the right build dependencies are setup and for
    the runtime it ensures the application is run in an environment catered
    for Flutter applications.

    It configures each application with the following plugs:

    \b
    - GTK3 Themes.
    - Common Icon Themes.
    - Common Sound Themes.
    - The GNOME runtime libraries and utilities corresponding to 3.28.

    For easier desktop integration, it also configures each application
    entry with these additional plugs:

    \b
    - desktop (https://snapcraft.io/docs/desktop-interface)
    - desktop-legacy (https://snapcraft.io/docs/desktop-legacy-interface)
    - gsettings (https://snapcraft.io/docs/gsettings-interface)
    - wayland (https://snapcraft.io/docs/wayland-interface)
    - opengl (https://snapcraft.io/docs/opengl-interface)
    - x11 (https://snapcraft.io/docs/x11-interface)
"""
)


def get_platform_snap_for_base(self, base: str) -> str:
    return _PLATFORM_SNAP[base]


class FlutterMetaExtension(type):
    def __new__(cls, name, bases, attrs):
        # Not setting these is a development error.
        channel = attrs["channel"]

        def is_experimental(base: Optional[str]) -> bool:
            return False

        def get_supported_bases() -> Tuple[str, ...]:
            return attrs["supported_bases"]

        def get_supported_confinement() -> Tuple[str, ...]:
            return attrs["supported_confinement"]

        def __repr__(self) -> str:
            return f"<Flutter Extension for {channel!r}>"

        def __init__(self, *, extension_name: str, yaml_data: Dict[str, Any]) -> None:
            base: Optional[str] = yaml_data.get("base")
            # A base is required in order to use extensions, so raise an error if not specified.
            if base is None:
                raise errors.ExtensionBaseRequiredError()

            if base not in self.get_supported_bases():
                raise errors.ExtensionUnsupportedBaseError(extension_name, base)

            # Default to devmode if confinement is not set.
            confinement = yaml_data.get("confinement", "devmode")
            if confinement not in self.get_supported_confinement():
                raise errors.ExtensionUnsupportedConfinementError(
                    extension_name, confinement
                )

            platform_snap = _PLATFORM_SNAP[base]
            self.root_snippet["plugs"].update(
                {
                    platform_snap: {
                        "interface": "content",
                        "target": "$SNAP/gnome-platform",
                        "default-provider": "{snap}".format(snap=platform_snap),
                    }
                }
            )

        x = super().__new__(
            cls,
            name,
            bases,
            {
                get_supported_bases.__name__: staticmethod(get_supported_bases),
                get_supported_confinement.__name__: staticmethod(
                    get_supported_confinement
                ),
                is_experimental.__name__: staticmethod(is_experimental),
                __repr__.__name__: __repr__,
                __init__.__name__: __init__,
            },
        )

        x.__doc__ = _DOC_TEMPLATE.format(channel=channel)
        x.root_snippet = {
            "assumes": ["snapd2.43"],  # for 'snapctl is-connected'
            "plugs": {
                "gtk-3-themes": {
                    "interface": "content",
                    "target": "$SNAP/data-dir/themes",
                    "default-provider": "gtk-common-themes",
                },
                "icon-themes": {
                    "interface": "content",
                    "target": "$SNAP/data-dir/icons",
                    "default-provider": "gtk-common-themes",
                },
                "sound-themes": {
                    "interface": "content",
                    "target": "$SNAP/data-dir/sounds",
                    "default-provider": "gtk-common-themes",
                },
            },
            "environment": {"SNAP_DESKTOP_RUNTIME": "$SNAP/gnome-platform"},
            "layout": {
                "/usr/share/xml/iso-codes": {
                    "bind": "$SNAP/gnome-platform/usr/share/xml/iso-codes"
                }
            },
        }

        x.app_snippet = {
            "command-chain": ["snap/command-chain/desktop-launch"],
            "plugs": [
                "desktop",
                "desktop-legacy",
                "gsettings",
                "opengl",
                "wayland",
                "x11",
            ],
        }

        x.part_snippet = dict()

        x.parts = {
            "gnome-3-28-extension": {
                "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
                "source-subdir": "gnome",
                "plugin": "make",
                "make-parameters": ["PLATFORM_PLUG=gnome-3-28-1804"],
                "build-packages": ["gcc", "libgtk-3-dev"],
            },
            "flutter-extension": {
                "plugin": "nil",
                "override-pull": textwrap.dedent(
                    f"""\
                    flutter channel {attrs["channel"]}
                    flutter config --enable-linux-desktop
                    flutter upgrade
                    flutter doctor
                    """
                ),
                "build-snaps": ["flutter/latest/stable"],
            },
        }

        return x



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict

import jsonschema

from ._api_schema import WHOAMI_JSONSCHEMA

"""
This module holds representations for results for the v2 whoami
API endpoint provided by the Snap Store.

The full API is documented on
https://dashboard.snapcraft.io/docs/v2/en/tokens.html#api-tokens-whoami
"""


class Account:
    """Represent Account information for a whoami response."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "Account":
        jsonschema.validate(payload, WHOAMI_JSONSCHEMA["properties"]["account"])
        return cls(
            email=payload["email"],
            account_id=payload["id"],
            name=payload["name"],
            username=payload["username"],
        )

    def marshal(self) -> Dict[str, Any]:
        return {
            "email": self.email,
            "id": self.account_id,
            "name": self.name,
            "username": self.username,
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.email!r}>"

    def __init__(
        self, *, email: str, account_id: str, name: str, username: str
    ) -> None:
        self.email = email
        self.account_id = account_id
        self.name = name
        self.username = username


class WhoAmI:
    """Represent the data returned from the whoami Snap Store endpoint."""

    @classmethod
    def unmarshal(cls, payload: Dict[str, Any]) -> "WhoAmI":
        jsonschema.validate(payload, WHOAMI_JSONSCHEMA)
        return cls(account=Account.unmarshal(payload["account"]))

    def marshal(self) -> Dict[str, Any]:
        return {
            "account": self.account.marshal(),
        }

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self.account.email!r}>"

    def __init__(self, *, account: Account) -> None:
        self.account = account



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Attributes that the jsonschema from the Snap Store originally require but
# Snapcraft does not have been commented out from this schema originally
# imported from
# https://dashboard.snapcraft.io/docs/v2/en/snaps.html#snap-channel-map

from typing import Any, Dict

# Version 14, found at: https://dashboard.snapcraft.io/docs/v2/en/snaps.html#snap-releases
RELEASES_JSONSCHEMA: Dict[str, Any] = {
    "properties": {
        "releases": {
            "items": {
                "properties": {
                    "architecture": {"introduced_at": 1, "type": "string"},
                    "branch": {"introduced_at": 1, "type": ["string", "null"]},
                    "channel": {
                        "description": "The channel name for this release.",
                        "introduced_at": 8,
                        "type": "string",
                    },
                    "expiration-date": {
                        "description": "The date when this release expires, in ISO 8601 format. If null, the release does not expire.",
                        "format": "date-time",
                        "introduced_at": 8,
                        "type": ["string", "null"],
                    },
                    "progressive": {
                        "introduced_at": 4,
                        "properties": {
                            "current-percentage": {
                                "introduced_at": 14,
                                "type": ["number", "null"],
                            },
                            "paused": {"type": ["boolean", "null"]},
                            "percentage": {"type": ["number", "null"]},
                        },
                        "required": ["paused", "percentage", "current-percentage"],
                        "type": "object",
                    },
                    "revision": {"introduced_at": 1, "type": ["integer", "null"]},
                    "risk": {"introduced_at": 1, "type": "string"},
                    "track": {"introduced_at": 1, "type": "string"},
                    "when": {
                        "format": "date-time",
                        "introduced_at": 1,
                        "type": "string",
                    },
                },
                "required": [
                    "architecture",
                    "branch",
                    "revision",
                    "risk",
                    "track",
                    "when",
                ],
                "type": "object",
            },
            "minItems": 0,
            "type": "array",
        },
        "revisions": {
            "items": {
                "properties": {
                    "architectures": {
                        "introduced_at": 1,
                        "items": {"type": "string"},
                        "minItems": 1,
                        "type": "array",
                    },
                    "attributes": {"introduced_at": 2, "type": "object"},
                    "base": {"introduced_at": 1, "type": ["string", "null"]},
                    "build_url": {"introduced_at": 1, "type": ["string", "null"]},
                    "confinement": {
                        "enum": ["strict", "classic", "devmode"],
                        "introduced_at": 1,
                        "type": "string",
                    },
                    "created_at": {
                        "format": "date-time",
                        "introduced_at": 1,
                        "type": "string",
                    },
                    "epoch": {"introduced_at": 1, "type": "object"},
                    "grade": {
                        "enum": ["stable", "devel"],
                        "introduced_at": 1,
                        "type": "string",
                    },
                    "revision": {"introduced_at": 1, "type": "integer"},
                    "sha3-384": {"introduced_at": 1, "type": "string"},
                    "size": {"introduced_at": 1, "type": "integer"},
                    "status": {
                        "enum": [
                            "Published",
                            "Unpublished",
                            "ManualReviewPending",
                            "NeedsInformation",
                            "AutomaticallyRejected",
                            "Rejected",
                        ],
                        "introduced_at": 1,
                        "type": "string",
                    },
                    "version": {"introduced_at": 1, "type": "string"},
                },
                "required": [
                    "architectures",
                    "build_url",
                    "confinement",
                    "created_at",
                    "grade",
                    "revision",
                    "sha3-384",
                    "size",
                    "status",
                    "version",
                ],
                "type": "object",
            },
            "minItems": 0,
            "type": "array",
        },
    },
    "required": ["releases", "revisions"],
    "type": "object",
}


CHANNEL_MAP_JSONSCHEMA: Dict[str, Any] = {
    "properties": {
        "channel-map": {
            "items": {
                "properties": {
                    "architecture": {"type": "string"},
                    "channel": {
                        "description": 'The channel name, including "latest/" for the latest track.',
                        "type": "string",
                    },
                    "expiration-date": {
                        "description": "The date when this release expires, in RFC 3339 format. If null, the release does not expire.",
                        "format": "date-time",
                        "type": ["string", "null"],
                    },
                    "progressive": {
                        "properties": {
                            "paused": {"type": ["boolean", "null"]},
                            "percentage": {"type": ["number", "null"]},
                            "current-percentage": {"type": ["number", "null"]},
                        },
                        "required": ["paused", "percentage", "current-percentage"],
                        "type": "object",
                    },
                    "revision": {"type": "integer"},
                    "when": {
                        "description": "The date when this release was made, in RFC 3339 format.",
                        "format": "date-time",
                        "type": "string",
                    },
                },
                "required": [
                    "architecture",
                    "channel",
                    "expiration-date",
                    "progressive",
                    "revision",
                    # "when"
                ],
                "type": "object",
            },
            "minItems": 0,
            "type": "array",
        },
        "revisions": {
            "items": {
                "properties": {
                    "architectures": {
                        "items": {"type": "string"},
                        "minItems": 1,
                        "type": "array",
                    },
                    "attributes": {"type": "object"},
                    "base": {"type": ["string", "null"]},
                    "build-url": {"type": ["string", "null"]},
                    "confinement": {
                        "enum": ["strict", "classic", "devmode"],
                        "type": "string",
                    },
                    "created-at": {"format": "date-time", "type": "string"},
                    "epoch": {
                        "properties": {
                            "read": {
                                "items": {"type": "integer"},
                                "minItems": 1,
                                "type": ["array", "null"],
                            },
                            "write": {
                                "items": {"type": "integer"},
                                "minItems": 1,
                                "type": ["array", "null"],
                            },
                        },
                        "required": ["read", "write"],
                        "type": "object",
                    },
                    "grade": {"enum": ["stable", "devel"], "type": "string"},
                    "revision": {"type": "integer"},
                    "sha3-384": {"type": "string"},
                    "size": {"type": "integer"},
                    "version": {"type": "string"},
                },
                "required": [
                    "architectures",
                    # "attributes",
                    # "base",
                    # "build-url",
                    # "confinement",
                    # "created-at",
                    # "epoch",
                    # "grade",
                    "revision",
                    # "sha3-384",
                    # "size",
                    # "status",
                    "version",
                ],
                "type": "object",
            },
            "minItems": 0,
            "type": "array",
        },
        "snap": {
            "description": "Metadata about the requested snap.",
            "introduced_at": 6,
            "properties": {
                "channels": {
                    "description": "The list of most relevant channels for this snap. Branches are only included if there is a release for it.",
                    "introduced_at": 9,
                    "items": {
                        "description": "A list of channels and their metadata for the requested snap.",
                        "properties": {
                            "branch": {
                                "description": "The branch name for this channel, can be null.",
                                "type": ["string", "null"],
                            },
                            "fallback": {
                                "description": "The name of the channel that this channel would fall back to if there were no releases in it. If null, this channel has no fallback channel.",
                                "type": ["string", "null"],
                            },
                            "name": {
                                "description": 'The channel name, including "latest/" for the latest track.',
                                "type": "string",
                            },
                            "risk": {
                                "description": "The risk name for this channel.",
                                "type": "string",
                            },
                            "track": {
                                "description": "The track name for this channel.",
                                "type": "string",
                            },
                        },
                        "required": ["name", "track", "risk", "branch", "fallback"],
                        "type": "object",
                    },
                    "minItems": 1,
                    "type": "array",
                },
                "default-track": {
                    "description": "The default track name for this snap. If no default track is set, this value is null.",
                    "type": ["string", "null"],
                },
                "id": {
                    "description": "The snap ID for this snap package.",
                    "type": "string",
                },
                "name": {"description": "The snap package name.", "type": "string"},
                "private": {
                    "description": "Whether this snap is private or not.",
                    "type": "boolean",
                },
                "tracks": {
                    "description": "An ordered list of most relevant tracks for this snap.",
                    "introduced_at": 9,
                    "items": {
                        "description": "An ordered list of tracks and their metadata for this snap.",
                        "properties": {
                            "creation-date": {
                                "description": "The track creation date, in ISO 8601 format.",
                                "format": "date-time",
                                "type": ["string", "null"],
                            },
                            "name": {
                                "description": "The track name.",
                                "type": "string",
                            },
                            "version-pattern": {
                                "description": "A Python regex to validate the versions being released to this track. If null, no validation is enforced.",
                                "type": ["string", "null"],
                            },
                        },
                        # pattern is documented as required but is not returned,
                        # version-pattern is returned instead.
                        "required": ["name", "creation-date", "version-pattern"],
                        "type": "object",
                    },
                    "minItems": 1,
                    "type": "array",
                },
            },
            "required": [
                # "id",
                "channels",
                # "default-track",
                "name",
                # "private",
                # "tracks"
            ],
            "type": "object",
        },
    },
    "required": ["channel-map", "revisions", "snap"],
    "type": "object",
}

# Version 27, found at https://dashboard.snapcraft.io/docs/v2/en/tokens.html#api-tokens-whoami
WHOAMI_JSONSCHEMA: Dict[str, Any] = {
    "properties": {
        "account": {
            "properties": {
                "email": {
                    "description": "The primary email for the user",
                    "type": "string",
                },
                "id": {
                    "description": "The store account ID for the user",
                    "type": "string",
                },
                "name": {
                    "description": "The display name for the user",
                    "type": "string",
                },
                "username": {
                    "description": "The store username for the user",
                    "type": "string",
                },
            },
            "required": ["email", "id", "name", "username"],
            "type": "object",
        },
        "channels": {
            "oneOf": [
                {
                    "description": "A list of channels to restrict the macaroon to, allows wildcards in the fnmatch format (https://docs.python.org/3/library/fnmatch.html)",
                    "items": {"description": "Valid channel name.", "type": "string"},
                    "minItems": 1,
                    "type": "array",
                    "uniqueItems": True,
                },
                {"type": "null"},
            ]
        },
        "errors": {
            "items": {
                "properties": {
                    "code": {"type": "string"},
                    "extra": {"additionalProperties": True, "type": "object"},
                    "message": {"type": "string"},
                },
                "required": ["code", "message", "extra"],
                "type": "object",
            },
            "type": "array",
        },
        "expires": {
            "oneOf": [
                {
                    "description": "A timestamp (string formatted per ISO8601) after which the macaroon should not be valid",
                    "format": "date-time",
                    "type": "string",
                },
                {"type": "null"},
            ]
        },
        "packages": {
            "oneOf": [
                {
                    "description": 'A list of packages to restrict the macaroon to. Those can be defined in two ways: 1- by name: each item in the list should be a json dict like `{"name": "the-name"}`, or 2- by snap_id: each item in the list should be a json dict like `{"snap_id": "some-snap-id-1234"}`.',
                    "items": {
                        "anyOf": [{"required": ["snap_id"]}, {"required": ["name"]}],
                        "description": "Package identifier: a dict with at least name or snap_id keys.",
                        "properties": {
                            "name": {"type": "string"},
                            "snap_id": {"type": "string"},
                        },
                        "type": "object",
                    },
                    "minItems": 1,
                    "type": "array",
                    "uniqueItems": True,
                },
                {"type": "null"},
            ]
        },
        "permissions": {
            "oneOf": [
                {
                    "description": "A list of permissions to restrict the macaroon to",
                    "items": {
                        "description": "A valid token permission.",
                        "enum": [
                            "edit_account",
                            "modify_account_key",
                            "package_access",
                            "package_manage",
                            "package_metrics",
                            "package_purchase",
                            "package_push",
                            "package_register",
                            "package_release",
                            "package_update",
                            "package_upload",
                            "package_upload_request",
                            "store_admin",
                            "store_review",
                        ],
                        "type": "string",
                    },
                    "minItems": 1,
                    "type": "array",
                    "uniqueItems": True,
                },
                {"type": "null"},
            ]
        },
        "store_ids": {
            "oneOf": [
                {
                    "description": "A list of store IDs to restrict the macaroon to",
                    "items": {"description": "A store ID.", "type": "string"},
                    "minItems": 1,
                    "type": "array",
                    "uniqueItems": True,
                },
                {"type": "null"},
            ]
        },
    },
    "required": ["account", "channels", "packages", "permissions"],
    "type": "object",
}

# https://dashboard.snapcraft.io/docs/v2/en/validation-sets.html
BUILD_ASSERTION_JSONSCHEMA: Dict[str, Any] = {
    "properties": {
        "account-id": {
            "description": 'The "account-id" assertion header',
            "type": "string",
        },
        "authority-id": {
            "description": 'The "authority-id" assertion header',
            "type": "string",
        },
        "name": {"description": 'The "name" assertion header', "type": "string"},
        "revision": {
            "description": 'The "revision" assertion header',
            "type": "string",
        },
        "sequence": {
            "description": 'The "sequence" assertion header',
            "type": "string",
        },
        "series": {"description": 'The "series" assertion header', "type": "string"},
        "snaps": {
            "items": {
                "description": "List of snaps in a Validation Set assertion",
                "properties": {
                    "id": {
                        "description": "Snap ID",
                        "maxLength": 100,
                        "type": "string",
                    },
                    "name": {
                        "description": "Snap name",
                        "maxLength": 100,
                        "type": "string",
                    },
                    "presence": {
                        "description": "Snap presence",
                        "enum": ["required", "optional", "invalid"],
                        "type": "string",
                    },
                    "revision": {"description": "Snap revision", "type": "string"},
                },
                "required": ["name"],
                "type": "object",
            },
            "minItems": 1,
            "type": "array",
        },
        "timestamp": {
            "description": 'The "timestamp" assertion header',
            "type": "string",
        },
        "type": {
            "const": "validation-set",
            "description": 'The "type" assertion header',
            "type": "string",
        },
    },
    "required": [
        "type",
        "authority-id",
        "series",
        "account-id",
        "name",
        "sequence",
        # "revision",
        "timestamp",
        "snaps",
    ],
    "type": "object",
}



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

from testtools.matchers import Equals, HasLength

import snapcraft
from snapcraft.internal import errors
from snapcraft.plugins.v1 import ruby

from . import PluginsV1BaseTestCase


class RubyPluginTestCase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        class Options(snapcraft.ProjectOptions):
            source = "."
            ruby_version = "2.4.2"
            gems = []
            use_bundler = False

        self.options = Options()

    def test_schema(self):
        schema = ruby.RubyPlugin.schema()
        expected_use_bundler = {"type": "boolean", "default": False}
        expected_ruby_version = {
            "type": "string",
            "default": "2.4.2",
            "pattern": r"^\d+\.\d+(\.\d+)?$",
        }
        expected_gems = {
            "type": "array",
            "minitems": 1,
            "uniqueItems": True,
            "items": {"type": "string"},
            "default": [],
        }

        self.assertThat(schema["properties"], HasLength(3))
        self.assertDictEqual(expected_use_bundler, schema["properties"]["use-bundler"])
        self.assertDictEqual(
            expected_ruby_version, schema["properties"]["ruby-version"]
        )
        self.assertDictEqual(expected_gems, schema["properties"]["gems"])

    def test_unsupported_base(self):
        self.project._snap_meta.base = "unsupported-base"

        raised = self.assertRaises(
            errors.PluginBaseError,
            ruby.RubyPlugin,
            "test-part",
            self.options,
            self.project,
        )

        self.assertThat(raised.part_name, Equals("test-part"))
        self.assertThat(raised.base, Equals("unsupported-base"))

    def test_get_pull_properties(self):
        expected_pull_properties = ["ruby-version", "gems", "use-bundler"]
        resulting_pull_properties = ruby.RubyPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_snap_fileset(self):
        expected_fileset = ["-include/", "-share/"]

        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        fileset = plugin.snap_fileset()

        self.assertThat(fileset, Equals(expected_fileset))

    def test_env_without_ruby(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        env = plugin.env("dummy-path")
        self.assertThat(env, Equals([]))

    def test_env_with_ruby(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        part_dir = os.path.join(self.path, "test-part-path")

        os.makedirs(os.path.join(part_dir, "lib", "ruby", "gems", "test-version"))
        libdir = os.path.join("test-part-path", "lib", "ruby", "test-version")
        arch_libdir = os.path.join(libdir, "foo-linux-bar")
        real_arch_libdir = os.path.join(self.path, arch_libdir)
        os.makedirs(real_arch_libdir)
        open(os.path.join(real_arch_libdir, "rbconfig.rb"), "w").close()
        env = plugin.env("test-part-path")

        expected_env = {
            'GEM_HOME="test-part-path/lib/ruby/gems/test-version"',
            'RUBYLIB="{}:{}"'.format(libdir, arch_libdir),
            'GEM_PATH="test-part-path/lib/ruby/gems/test-version"',
        }
        self.assertThat(set(env), Equals(expected_env))

    def test_env_with_multiple_ruby(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        part_dir = os.path.join(self.path, "test-part-path")

        os.makedirs(os.path.join(part_dir, "lib", "ruby", "gems", "test-version1"))
        os.makedirs(os.path.join(part_dir, "lib", "ruby", "gems", "test-version2"))

        error = self.assertRaises(
            snapcraft.internal.errors.SnapcraftEnvironmentError,
            plugin.env,
            "test-part-path",
        )

        self.assertThat(
            str(error), Equals("Expected a single Ruby version, but found 2")
        )

    def test_env_with_rbconfigs(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        part_dir = os.path.join(self.path, "test-part-path")

        os.makedirs(os.path.join(part_dir, "lib", "ruby", "gems", "test-version"))
        libdir = os.path.join("test-part-path", "lib", "ruby", "test-version")

        for arch in ("foo-linux-bar1", "foo-linux-bar2"):
            arch_libdir1 = os.path.join(libdir, arch)
            real_arch_libdir1 = os.path.join(self.path, arch_libdir1)
            os.makedirs(real_arch_libdir1)
            open(os.path.join(real_arch_libdir1, "rbconfig.rb"), "w").close()

        error = self.assertRaises(
            snapcraft.internal.errors.SnapcraftEnvironmentError,
            plugin.env,
            "test-part-path",
        )

        self.assertThat(
            str(error), Equals("Expected a single rbconfig.rb, but found 2")
        )

    def test_pull_downloads_ruby(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)

        self.assertThat(
            plugin._ruby_tar.source,
            Equals("https://cache.ruby-lang.org/pub/ruby/2.4/ruby-2.4.2.tar.gz"),
        )
        self.assertThat(
            plugin._ruby_tar.source_dir,
            Equals(os.path.join(self.path, "parts", "test-part", "ruby")),
        )

        with mock.patch.multiple(
            plugin, _ruby_install=mock.DEFAULT, _gem_install=mock.DEFAULT
        ):
            # We don't want to wait for the download to finish. It is already
            # tested in the Tar object tests.
            with mock.patch.object(plugin._ruby_tar, "download") as mock_download:
                plugin.pull()

        mock_download.assert_called_once_with()

    def test_pull_installs_ruby(self):
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        with mock.patch.multiple(
            plugin, _ruby_tar=mock.DEFAULT, _gem_install=mock.DEFAULT
        ) as mocks:
            with mock.patch("snapcraft.internal.common.run") as mock_run:
                plugin.pull()

        ruby_expected_dir = os.path.join(self.path, "parts", "test-part", "ruby")
        mocks["_ruby_tar"].provision.assert_called_with(
            ruby_expected_dir, clean_target=False, keep_tarball=True
        )
        mock_run.assert_has_calls(
            [
                mock.call(
                    ["./configure", "--disable-install-rdoc", "--prefix=/"],
                    cwd=ruby_expected_dir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["make", "-j{}".format(plugin.parallel_build_count)],
                    cwd=ruby_expected_dir,
                    env=mock.ANY,
                ),
                mock.call(
                    ["make", "install", "DESTDIR={}".format(plugin.installdir)],
                    cwd=ruby_expected_dir,
                    env=mock.ANY,
                ),
            ]
        )

    def test_pull_installs_gems_without_bundler(self):
        self.options.gems = ["test-gem-1", "test-gem-2"]
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        with mock.patch.multiple(
            plugin, _ruby_tar=mock.DEFAULT, _ruby_install=mock.DEFAULT
        ):
            with mock.patch("snapcraft.internal.common.run") as mock_run:
                plugin.pull()

        test_part_dir = os.path.join(self.path, "parts", "test-part")
        mock_run.assert_called_with(
            [
                os.path.join(test_part_dir, "install", "bin", "ruby"),
                os.path.join(test_part_dir, "install", "bin", "gem"),
                "install",
                "--env-shebang",
                "test-gem-1",
                "test-gem-2",
            ],
            cwd=os.path.join(test_part_dir, "build"),
            env=mock.ANY,
        )

    def test_pull_with_bundler(self):
        self.options.gems = ["test-gem-1", "test-gem-2"]
        plugin = ruby.RubyPlugin("test-part", self.options, self.project)
        self.options.use_bundler = True

        with mock.patch.multiple(
            plugin, _ruby_tar=mock.DEFAULT, _ruby_install=mock.DEFAULT
        ):
            with mock.patch("snapcraft.internal.common.run") as mock_run:
                plugin.pull()
        test_part_dir = os.path.join(self.path, "parts", "test-part")
        mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(test_part_dir, "install", "bin", "ruby"),
                        os.path.join(test_part_dir, "install", "bin", "gem"),
                        "install",
                        "--env-shebang",
                        "test-gem-1",
                        "test-gem-2",
                        "bundler",
                    ],
                    cwd=os.path.join(test_part_dir, "build"),
                    env=mock.ANY,
                ),
                mock.call(
                    [
                        os.path.join(test_part_dir, "install", "bin", "ruby"),
                        os.path.join(test_part_dir, "install", "bin", "bundle"),
                        "install",
                    ],
                    cwd=os.path.join(test_part_dir, "build"),
                    env=mock.ANY,
                ),
            ]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import textwrap
from unittest.mock import Mock

import pytest
import tinydb

from snapcraft.internal.db import datastore, errors, migration


@pytest.fixture(autouse=True)
def mock_datetime(monkeypatch):
    datetime_mock = Mock(wraps=datetime.datetime)
    datetime_mock.utcnow.return_value = datetime.datetime(2020, 1, 2, 3, 4, 5, 6)
    monkeypatch.setattr(migration, "datetime", datetime_mock)


def test_open_invalid_data(tmp_path):
    db_path = tmp_path / "db.yaml"
    db_path.write_text("invalid")

    with pytest.raises(RuntimeError) as error:
        with datastore.Datastore(path=db_path, migrations=[]) as _:
            assert False

    assert (
        error.value.args[0]
        == f"Invalid datastore contents for {str(db_path)}: 'invalid'"
    )


def test_create_no_migration(tmp_path):
    db_path = tmp_path / "db.yaml"

    with datastore.Datastore(path=db_path, migrations=[], snapcraft_version="42") as db:
        assert isinstance(db, tinydb.TinyDB) is True
        db.table("test").insert({"foo": "bar"})

    assert db_path.read_text() == textwrap.dedent(
        """\
        test:
          '1':
            foo: bar
        """
    )


def test_create_default_migration(tmp_path):
    db_path = tmp_path / "db.yaml"

    with datastore.Datastore(
        path=db_path, migrations=[migration.MigrationV1], snapcraft_version="42"
    ) as db:
        assert isinstance(db, tinydb.TinyDB) is True
        db.table("test").insert({"foo": "bar"})

    assert db_path.read_text() == textwrap.dedent(
        """\
        control:
          '1':
            created_with_snapcraft_version: '42'
            schema_version: 1
        migration:
          '1':
            schema_version: 1
            snapcraft_version: '42'
            timestamp: '2020-01-02T03:04:05.000006Z'
        test:
          '1':
            foo: bar
        """
    )


def test_unknown_version(tmp_path):
    db_path = tmp_path / "db.yaml"
    control_record = {"created_with_snapcraft_version": "42", "schema_version": 3}

    # Populate test db with schema version greater than 1.
    with datastore.Datastore(path=db_path, migrations=[], snapcraft_version="42") as db:
        assert isinstance(db, tinydb.TinyDB) is True
        db.table("control").insert(control_record)

    # Re-open db, should raise version unsupported error.
    with pytest.raises(errors.SnapcraftDatastoreVersionUnsupported):
        with datastore.Datastore(
            path=db_path, migrations=[migration.MigrationV1]
        ) as db:
            assert False

    # Verify we are able to open unsupported version in read-only.
    with datastore.Datastore(
        path=db_path, migrations=[migration.MigrationV1], read_only=True
    ) as db:
        assert db.table("control").all() == [control_record]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys
from unittest import mock

from testtools.matchers import DirExists, Equals, FileExists, MatchesRegex

from snapcraft.internal import sources
from tests import unit


class TestSnap(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.test_file_path = os.path.join(
            os.path.dirname(unit.__file__), "..", "data", "test-snap.snap"
        )

        self.dest_dir = "dst"
        os.makedirs(self.dest_dir)

    def test_pull_snap_file_must_extract(self):
        snap_source = sources.Snap(self.test_file_path, self.dest_dir)
        snap_source.pull()

        self.assertThat(os.path.join(self.dest_dir, "meta.basic"), DirExists())
        self.assertThat(
            os.path.join(self.dest_dir, "meta.basic", "snap.yaml"), FileExists()
        )

    @mock.patch.object(sources.Snap, "provision")
    def test_pull_snap_must_not_clean_targets(self, mock_provision):
        snap_source = sources.Snap(self.test_file_path, self.dest_dir)
        snap_source.pull()

        mock_provision.assert_called_once_with(
            self.dest_dir,
            clean_target=False,
            src=os.path.join(self.dest_dir, "test-snap.snap"),
        )

    def test_has_source_handler_entry_on_linux(self):
        if sys.platform == "linux":
            self.assertTrue(sources._source_handler["snap"] is sources.Snap)
        else:
            self.assertRaises(KeyError, sources._source_handler["snap"])

    @mock.patch(
        "subprocess.check_output", side_effect=subprocess.CalledProcessError(1, [])
    )
    def test_pull_failure_bad_unsquash(self, mock_run):
        snap_source = sources.Snap(self.test_file_path, self.dest_dir)
        raised = self.assertRaises(sources.errors.SnapcraftPullError, snap_source.pull)
        self.assertThat(
            raised.command,
            MatchesRegex(
                r"unsquashfs -force -dest {0}/\w+ {0}/dst/test-snap.snap".format(
                    self.path
                )
            ),
        )
        self.assertThat(raised.exit_code, Equals(1))


class TestGetName(unit.TestCase):
    def test_get_name(self):
        os.mkdir("meta")
        with open(os.path.join("meta", "snap.yaml"), "w") as snap_yaml_file:
            print("name: my-snap", file=snap_yaml_file)
        self.assertThat(sources._snap._get_snap_name("."), Equals("my-snap"))

    def test_no_name_yaml(self):
        os.mkdir("meta")
        with open(os.path.join("meta", "snap.yaml"), "w") as snap_yaml_file:
            print("summary: no name", file=snap_yaml_file)
        self.assertRaises(
            sources.errors.InvalidSnapError, sources._snap._get_snap_name, "."
        )

    def test_no_snap_yaml(self):
        os.mkdir("meta")
        self.assertRaises(
            sources.errors.InvalidSnapError, sources._snap._get_snap_name, "."
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os

from ._cache import SnapcraftStagePackageCache

logger = logging.getLogger(__name__)


class AptStagePackageCache(SnapcraftStagePackageCache):
    """Cache for stage-packages coming from apt."""

    def __init__(self, *, sources_digest):
        """Create a new AptStagePackageCache.

        :param str sources_digest: Unique digest of the current apt sources.
        """

        super().__init__()
        cache_base_dir = os.path.join(self.stage_package_cache_root, "apt")

        # TODO: Clean old cache, LP: #1663051

        self.base_dir = os.path.join(cache_base_dir, sources_digest)
        self.packages_dir = os.path.join(
            self.base_dir, "var", "cache", "apt", "archives"
        )
        os.makedirs(self.packages_dir, exist_ok=True)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pathlib
import subprocess
from textwrap import dedent
from unittest import mock
from unittest.mock import call

import pytest

from snapcraft.internal.meta.package_repository import (
    PackageRepositoryApt,
    PackageRepositoryAptPpa,
)
from snapcraft.internal.repo import apt_ppa, apt_sources_manager, errors


@pytest.fixture(autouse=True)
def mock_apt_ppa_get_signing_key():
    with mock.patch(
        "snapcraft.internal.repo.apt_ppa.get_launchpad_ppa_key_id",
        spec=apt_ppa.get_launchpad_ppa_key_id,
        return_value="FAKE-PPA-SIGNING-KEY",
    ) as m:
        yield m


@pytest.fixture(autouse=True)
def mock_environ_copy():
    with mock.patch("os.environ.copy") as m:
        yield m


@pytest.fixture(autouse=True)
def mock_host_arch():
    with mock.patch("snapcraft.internal.repo.apt_sources_manager.ProjectOptions") as m:
        m.return_value.deb_arch = "FAKE-HOST-ARCH"
        yield m


@pytest.fixture(autouse=True)
def mock_run():
    with mock.patch("subprocess.run") as m:
        yield m


@pytest.fixture()
def mock_sudo_write():
    def write_file(*, dst_path: pathlib.Path, content: bytes) -> None:
        dst_path.write_bytes(content)

    with mock.patch(
        "snapcraft.internal.repo.apt_sources_manager._sudo_write_file"
    ) as m:
        m.side_effect = write_file
        yield m


@pytest.fixture(autouse=True)
def mock_version_codename():
    with mock.patch(
        "snapcraft.internal.os_release.OsRelease.version_codename",
        return_value="FAKE-CODENAME",
    ) as m:
        yield m


@pytest.fixture
def apt_sources_mgr(tmp_path):
    sources_list_d = tmp_path / "sources.list.d"
    sources_list_d.mkdir(parents=True)

    yield apt_sources_manager.AptSourcesManager(sources_list_d=sources_list_d,)


@mock.patch("tempfile.NamedTemporaryFile")
@mock.patch("os.unlink")
def test_sudo_write_file(mock_unlink, mock_tempfile, mock_run, tmp_path):
    mock_tempfile.return_value.__enter__.return_value.name = "/tmp/foobar"

    apt_sources_manager._sudo_write_file(dst_path="/foo/bar", content=b"some-content")

    assert mock_tempfile.mock_calls == [
        call(delete=False),
        call().__enter__(),
        call().__enter__().write(b"some-content"),
        call().__enter__().flush(),
        call().__exit__(None, None, None),
    ]
    assert mock_run.mock_calls == [
        call(
            [
                "sudo",
                "install",
                "--owner=root",
                "--group=root",
                "--mode=0644",
                "/tmp/foobar",
                "/foo/bar",
            ],
            check=True,
        )
    ]
    assert mock_unlink.mock_calls == [call("/tmp/foobar")]


def test_sudo_write_file_fails(mock_run):
    mock_run.side_effect = subprocess.CalledProcessError(
        cmd=["sudo"], returncode=1, output=b"some error"
    )

    with pytest.raises(RuntimeError) as error:
        apt_sources_manager._sudo_write_file(
            dst_path="/foo/bar", content=b"some-content"
        )

    assert (
        str(error.value).startswith(
            "Failed to install repository config with: ['sudo', 'install'"
        )
        is True
    )


@pytest.mark.parametrize(
    "package_repo,name,content",
    [
        (
            PackageRepositoryApt(
                architectures=["amd64", "arm64"],
                components=["test-component"],
                formats=["deb", "deb-src"],
                key_id="A" * 40,
                suites=["test-suite1", "test-suite2"],
                url="http://test.url/ubuntu",
            ),
            "snapcraft-http_test_url_ubuntu.sources",
            dedent(
                """\
                Types: deb deb-src
                URIs: http://test.url/ubuntu
                Suites: test-suite1 test-suite2
                Components: test-component
                Architectures: amd64 arm64
                """
            ).encode(),
        ),
        (
            PackageRepositoryApt(
                architectures=["amd64", "arm64"],
                components=["test-component"],
                key_id="A" * 40,
                name="NO-FORMAT",
                suites=["test-suite1", "test-suite2"],
                url="http://test.url/ubuntu",
            ),
            "snapcraft-NO-FORMAT.sources",
            dedent(
                """\
                Types: deb
                URIs: http://test.url/ubuntu
                Suites: test-suite1 test-suite2
                Components: test-component
                Architectures: amd64 arm64
                """
            ).encode(),
        ),
        (
            PackageRepositoryApt(
                key_id="A" * 40,
                name="WITH-PATH",
                path="some-path",
                url="http://test.url/ubuntu",
            ),
            "snapcraft-WITH-PATH.sources",
            dedent(
                """\
                Types: deb
                URIs: http://test.url/ubuntu
                Suites: some-path/
                Architectures: FAKE-HOST-ARCH
                """
            ).encode(),
        ),
        (
            PackageRepositoryApt(
                key_id="A" * 40, name="IMPLIED-PATH", url="http://test.url/ubuntu",
            ),
            "snapcraft-IMPLIED-PATH.sources",
            dedent(
                """\
                Types: deb
                URIs: http://test.url/ubuntu
                Suites: /
                Architectures: FAKE-HOST-ARCH
                """
            ).encode(),
        ),
        (
            PackageRepositoryAptPpa(ppa="test/ppa"),
            "snapcraft-ppa-test_ppa.sources",
            dedent(
                """\
                Types: deb
                URIs: http://ppa.launchpad.net/test/ppa/ubuntu
                Suites: FAKE-CODENAME
                Components: main
                Architectures: FAKE-HOST-ARCH
                """
            ).encode(),
        ),
    ],
)
def test_install(package_repo, name, content, apt_sources_mgr, mock_sudo_write):
    sources_path = apt_sources_mgr._sources_list_d / name

    changed = apt_sources_mgr.install_package_repository_sources(
        package_repo=package_repo
    )

    assert changed is True
    assert sources_path.read_bytes() == content
    assert mock_sudo_write.mock_calls == [call(content=content, dst_path=sources_path,)]

    # Verify a second-run does not incur any changes.
    mock_sudo_write.reset_mock()

    changed = apt_sources_mgr.install_package_repository_sources(
        package_repo=package_repo
    )

    assert changed is False
    assert sources_path.read_bytes() == content
    assert mock_sudo_write.mock_calls == []


def test_install_ppa_invalid(apt_sources_mgr):
    repo = PackageRepositoryAptPpa(ppa="ppa-missing-slash")

    with pytest.raises(errors.AptPPAInstallError) as exc_info:
        apt_sources_mgr.install_package_repository_sources(package_repo=repo)

    assert exc_info.value._ppa == "ppa-missing-slash"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Import types and tell flake8 to ignore the "unused" List.

from collections import namedtuple
from typing import Any, Dict, Optional, Tuple

from ._extension import Extension


_ExtensionInfo = namedtuple("ExtensionInfo", "cmake_args content provider build_snaps")

_Info = dict(
    core18=_ExtensionInfo(
        cmake_args=None,
        content="kde-frameworks-5-core18-all",
        provider="kde-frameworks-5-core18",
        build_snaps=["kde-frameworks-5-core18-sdk/latest/stable"],
    ),
    core20=_ExtensionInfo(
        cmake_args="-DCMAKE_FIND_ROOT_PATH=/snap/kde-frameworks-5-qt-5-15-core20-sdk/current",
        content="kde-frameworks-5-qt-5-15-core20-all",
        provider="kde-frameworks-5-qt-5-15-core20",
        build_snaps=["kde-frameworks-5-qt-5-15-core20-sdk/latest/candidate"],
    ),
)


class ExtensionImpl(Extension):
    """The KDE Neon extension.

    This extension makes it easy to assemble KDE based applications
    using the Neon stack.

    It configures each application with the following plugs:

    \b
    - Common Icon Themes.
    - Common Sound Themes.
    - The Qt5 and KDE Frameworks runtime libraries and utilities.

    For easier desktop integration, it also configures each application
    entry with these additional plugs:

    \b
    - desktop (https://snapcraft.io/docs/desktop-interface)
    - desktop-legacy (https://snapcraft.io/docs/desktop-legacy-interface)
    - opengl (https://snapcraft.io/docs/opengl-interface)
    - wayland (https://snapcraft.io/docs/wayland-interface)
    - x11 (https://snapcraft.io/docs/x11-interface)
    """

    @staticmethod
    def is_experimental(base: Optional[str]) -> bool:
        # TODO: remove experimental once sdk is on stable
        return base == "core20"

    @staticmethod
    def get_supported_bases() -> Tuple[str, ...]:
        return ("core18", "core20")

    @staticmethod
    def get_supported_confinement() -> Tuple[str, ...]:
        return ("strict", "devmode")

    def __init__(self, *, extension_name: str, yaml_data: Dict[str, Any]) -> None:
        super().__init__(extension_name=extension_name, yaml_data=yaml_data)

        info = _Info[yaml_data["base"]]
        self.root_snippet = {
            "assumes": ["snapd2.43"],  # for 'snapctl is-connected'
            "plugs": {
                "desktop": {"mount-host-font-cache": False},
                "icon-themes": {
                    "interface": "content",
                    "target": "$SNAP/data-dir/icons",
                    "default-provider": "gtk-common-themes",
                },
                "sound-themes": {
                    "interface": "content",
                    "target": "$SNAP/data-dir/sounds",
                    "default-provider": "gtk-common-themes",
                },
                "kde-frameworks-5-plug": {
                    "content": info.content,
                    "interface": "content",
                    "default-provider": info.provider,
                    "target": "$SNAP/kf5",
                },
            },
            "environment": {"SNAP_DESKTOP_RUNTIME": "$SNAP/kf5"},
            "hooks": {
                "configure": {
                    "plugs": ["desktop"],
                    "command-chain": ["snap/command-chain/hooks-configure-desktop"],
                }
            },
        }

        if info.cmake_args is not None:
            self.part_snippet = {
                "build-environment": [{"SNAPCRAFT_CMAKE_ARGS": info.cmake_args}]
            }

        self.app_snippet = {
            "command-chain": ["snap/command-chain/desktop-launch"],
            "plugs": ["desktop", "desktop-legacy", "opengl", "wayland", "x11"],
        }

        self.parts = {
            "kde-neon-extension": {
                "source": "$SNAPCRAFT_EXTENSIONS_DIR/desktop",
                "source-subdir": "kde-neon",
                "plugin": "make",
                "make-parameters": ["PLATFORM_PLUG=kde-frameworks-5-plug"],
                "build-packages": ["g++"],
                "build-snaps": info.build_snaps,
            }
        }



#!/usr/bin/env python3

import pytest

from snapcraft.internal import states


class Project:
    def __init__(self):
        self.deb_arch = "amd64"


_DEFAULT = (Project(), ["foo"], {"foo": "bar"})
_VARIANTS = [
    (Project(), [], {"foo": "bar"}),
    (Project(), ["foo"], None),
    (None, ["foo"], {"foo": "bar"}),
]


@pytest.fixture
def pull_state():
    """Return a PullState."""
    project, property_names, part_properties = _DEFAULT

    return states.PullState(property_names, part_properties, project)


@pytest.fixture(params=_VARIANTS)
def pull_state_variant(request):
    """Return variants of pull_state."""
    project, property_names, part_properties = request.param

    return states.PullState(property_names, part_properties, project)


@pytest.fixture
def build_state():
    """Return a BuildState."""
    project, property_names, part_properties = _DEFAULT

    return states.BuildState(property_names, part_properties, project)


@pytest.fixture(params=_VARIANTS)
def build_state_variant(request):
    """Return variants of build_state."""
    project, property_names, part_properties = request.param

    return states.BuildState(property_names, part_properties, project)


_STAGE_DEFAULT = (
    Project(),
    {"foo"},
    {"bar"},
    {
        "filesets": {"qux": "quux"},
        "override-stage": "touch override-stage",
        "stage": ["baz"],
    },
)
_STAGE_VARIANTS = [
    (
        Project(),
        set(),
        {"bar"},
        {
            "filesets": {"qux": "quux"},
            "override-stage": "touch override-stage",
            "stage": ["baz"],
        },
    ),
    (
        Project(),
        {"foo"},
        set(),
        {
            "filesets": {"qux": "quux"},
            "override-stage": "touch override-stage",
            "stage": ["baz"],
        },
    ),
    (Project(), {"foo"}, {"bar"}, None),
]


@pytest.fixture
def stage_state():
    """Return a StageState."""
    project, files, directories, part_properties = _STAGE_DEFAULT

    return states.StageState(files, directories, part_properties, project)


@pytest.fixture(params=_STAGE_VARIANTS)
def stage_state_variant(request):
    """Return variants of stage_state."""
    project, files, directories, part_properties = request.param

    return states.StageState(files, directories, part_properties, project)


_PRIME_DEFAULT = (
    Project(),
    {"foo"},
    {"bar"},
    {"baz"},
    {"override-prime": "touch override-prime", "prime": ["qux"]},
)
_PRIME_VARIANTS = [
    (
        Project(),
        set(),
        {"bar"},
        {"baz"},
        {"override-prime": "touch override-prime", "prime": ["qux"]},
    ),
    (
        Project(),
        {"foo"},
        {"bar"},
        set(),
        {"override-prime": "touch override-prime", "prime": ["qux"]},
    ),
    (Project(), {"foo"}, {"bar"}, {"baz"}, None),
]


@pytest.fixture
def prime_state():
    """Return a PrimeState."""
    project, files, directories, dependency_paths, part_properties = _PRIME_DEFAULT

    return states.PrimeState(
        files, directories, dependency_paths, part_properties, project
    )


@pytest.fixture(params=_PRIME_VARIANTS)
def prime_state_variant(request):
    """Return variants of prime_state."""
    project, files, directories, dependency_paths, part_properties = request.param

    return states.PrimeState(
        files, directories, dependency_paths, part_properties, project
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.storeapi.v2 import validation_sets


@pytest.mark.parametrize("snap_id", (None, "snap_id"))
@pytest.mark.parametrize("presence", (None, "required", "optional", "invalid"))
@pytest.mark.parametrize("revision", ("42", None))
def test_snap(snap_id, presence, revision):
    payload = {
        "name": "set-1",
    }

    if presence is not None:
        payload["presence"] = presence

    if snap_id is not None:
        payload["id"] = snap_id

    if revision is not None:
        payload["revision"] = revision

    s = validation_sets.Snap.unmarshal(payload)

    assert repr(s) == "<Snap: 'set-1'>"
    assert s.name == payload["name"]
    assert s.snap_id == payload.get("id")
    assert s.presence == payload.get("presence")
    assert s.revision == payload.get("revision")
    assert s.marshal() == payload


@pytest.mark.parametrize("revision", ("42", None))
def test_build_assertion(revision):
    payload = {
        "account-id": "account-id-1",
        "authority-id": "authority-id-1",
        "name": "validation-set-1",
        "sequence": "1",
        "series": "16",
        "snaps": [validation_sets.Snap(name="snap-name").marshal()],
        "timestamp": "2020-10-29T16:36:56Z",
        "type": "validation-set",
    }

    if revision is not None:
        payload["revision"] = revision

    s = validation_sets.BuildAssertion.unmarshal(payload)

    assert repr(s) == "<BuildAssertion: 'validation-set-1'>"
    assert s.account_id == payload["account-id"]
    assert s.authority_id == payload["authority-id"]
    assert s.name == payload["name"]
    assert s.sequence == payload.get("sequence")
    assert s.series == payload.get("series")
    assert s.snaps == [validation_sets.Snap(name="snap-name")]
    assert s.timestamp == "2020-10-29T16:36:56Z"

    if revision is None:
        payload["revision"] = "0"

    assert s.marshal() == payload


def test_validation_sets():
    payload = {
        "assertions": [
            {
                "headers": {
                    "account-id": "account-id-1",
                    "authority-id": "authority-id-1",
                    "name": "validation-set-1",
                    "revision": "1",
                    "sequence": "1",
                    "series": "16",
                    "snaps": [validation_sets.Snap(name="snap-name-1").marshal()],
                    "timestamp": "2020-10-29T16:36:56Z",
                    "type": "validation-set",
                }
            },
            {
                "headers": {
                    "account-id": "account-id-1",
                    "authority-id": "authority-id-1",
                    "name": "validation-set-1",
                    "revision": "3",
                    "sequence": "1",
                    "series": "16",
                    "snaps": [validation_sets.Snap(name="snap-name-2").marshal()],
                    "timestamp": "2020-10-29T16:36:56Z",
                    "type": "validation-set",
                },
            },
        ]
    }

    s = validation_sets.ValidationSets.unmarshal(payload)

    assert repr(s) == "<ValidationSets: assertions 2>"
    assert s.assertions[0] == validation_sets.BuildAssertion.unmarshal(
        {
            "account-id": "account-id-1",
            "authority-id": "authority-id-1",
            "name": "validation-set-1",
            "revision": "1",
            "sequence": "1",
            "series": "16",
            "snaps": [validation_sets.Snap(name="snap-name-1").marshal()],
            "timestamp": "2020-10-29T16:36:56Z",
            "type": "validation-set",
        },
    )

    assert s.assertions[1] == validation_sets.BuildAssertion.unmarshal(
        {
            "account-id": "account-id-1",
            "authority-id": "authority-id-1",
            "name": "validation-set-1",
            "revision": "3",
            "sequence": "1",
            "series": "16",
            "snaps": [validation_sets.Snap(name="snap-name-2").marshal()],
            "timestamp": "2020-10-29T16:36:56Z",
            "type": "validation-set",
        }
    )

    assert s.marshal() == payload



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from snapcraft.plugins.v2.python import PythonPlugin


def test_schema():
    assert PythonPlugin.get_schema() == {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "type": "object",
        "additionalProperties": False,
        "properties": {
            "constraints": {
                "default": [],
                "items": {"type": "string"},
                "type": "array",
                "uniqueItems": True,
            },
            "python-packages": {
                "default": ["pip", "setuptools", "wheel"],
                "items": {"type": "string"},
                "type": "array",
                "uniqueItems": True,
            },
            "requirements": {
                "default": [],
                "items": {"type": "string"},
                "type": "array",
                "uniqueItems": True,
            },
        },
        "type": "object",
    }


def test_get_build_packages():
    plugin = PythonPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_packages() == {"findutils", "python3-venv", "python3-dev"}


def test_get_build_environment():
    plugin = PythonPlugin(part_name="my-part", options=lambda: None)

    assert plugin.get_build_environment() == {
        "PATH": "${SNAPCRAFT_PART_INSTALL}/bin:${PATH}",
        "SNAPCRAFT_PYTHON_INTERPRETER": "python3",
        "SNAPCRAFT_PYTHON_VENV_ARGS": "",
    }


_FIXUP_BUILD_COMMANDS = [
    dedent(
        """\
            find "${SNAPCRAFT_PART_INSTALL}" -type f -executable -print0 | xargs -0 \
                sed -i "1 s|^#\\!${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}.*$|#\\!/usr/bin/env ${SNAPCRAFT_PYTHON_INTERPRETER}|"
            """
    ),
    dedent(
        """\
            determine_link_target() {
                opts_state="$(set +o +x | grep xtrace)"
                interp_dir="$(dirname "${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}")"
                # Determine python based on PATH, then resolve it, e.g:
                # (1) /home/ubuntu/.venv/snapcraft/bin/python3 -> /usr/bin/python3.8
                # (2) /usr/bin/python3 -> /usr/bin/python3.8
                # (3) /root/stage/python3 -> /root/stage/python3.8
                # (4) /root/parts/<part>/install/usr/bin/python3 -> /root/parts/<part>/install/usr/bin/python3.8
                python_path="$(which "${SNAPCRAFT_PYTHON_INTERPRETER}")"
                python_path="$(readlink -e "${python_path}")"
                for dir in "${SNAPCRAFT_PART_INSTALL}" "${SNAPCRAFT_STAGE}"; do
                    if  echo "${python_path}" | grep -q "${dir}"; then
                        python_path="$(realpath --strip --relative-to="${interp_dir}" \\
                                "${python_path}")"
                        break
                    fi
                done
                echo "${python_path}"
                eval "${opts_state}"
            }

            python_path="$(determine_link_target)"
            ln -sf "${python_path}" "${SNAPCRAFT_PYTHON_VENV_INTERP_PATH}"
        """
    ),
]


def test_get_build_commands():
    class Options:
        constraints = list()
        requirements = list()
        python_packages = list()

    plugin = PythonPlugin(part_name="my-part", options=Options())

    assert (
        plugin.get_build_commands()
        == [
            '"${SNAPCRAFT_PYTHON_INTERPRETER}" -m venv ${SNAPCRAFT_PYTHON_VENV_ARGS} "${SNAPCRAFT_PART_INSTALL}"',
            'SNAPCRAFT_PYTHON_VENV_INTERP_PATH="${SNAPCRAFT_PART_INSTALL}/bin/${SNAPCRAFT_PYTHON_INTERPRETER}"',
            "[ -f setup.py ] && pip install  -U .",
        ]
        + _FIXUP_BUILD_COMMANDS
    )


def test_get_build_commands_with_all_properties():
    class Options:
        constraints = ["constraints.txt"]
        requirements = ["requirements.txt"]
        python_packages = ["pip", "some-pkg; sys_platform != 'win32'"]

    plugin = PythonPlugin(part_name="my-part", options=Options())

    assert (
        plugin.get_build_commands()
        == [
            '"${SNAPCRAFT_PYTHON_INTERPRETER}" -m venv ${SNAPCRAFT_PYTHON_VENV_ARGS} "${SNAPCRAFT_PART_INSTALL}"',
            'SNAPCRAFT_PYTHON_VENV_INTERP_PATH="${SNAPCRAFT_PART_INSTALL}/bin/${SNAPCRAFT_PYTHON_INTERPRETER}"',
            "pip install -c 'constraints.txt' -U pip 'some-pkg; sys_platform != '\"'\"'win32'\"'\"''",
            "pip install -c 'constraints.txt' -U -r 'requirements.txt'",
            "[ -f setup.py ] && pip install -c 'constraints.txt' -U .",
        ]
        + _FIXUP_BUILD_COMMANDS
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The dotnet plugin is used to build dotnet core runtime parts.'

The plugin uses the dotnet SDK to install dependencies from nuget
and follows standard semantics from a dotnet core project.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

The plugin will take into account the following build-attributes:

    - debug: builds using a Debug configuration.

"""

import fnmatch
import json
import os
import shutil
import urllib.parse
import urllib.request
from typing import List

from snapcraft import formatting_utils, sources
from snapcraft.internal import errors
from snapcraft.plugins.v1 import PluginV1

_DOTNET_RELEASE_METADATA_URL = "https://dotnetcli.blob.core.windows.net/dotnet/release-metadata/{version}/releases.json"  # noqa
_RUNTIME_DEFAULT = "2.0.9"
_VERSION_DEFAULT = "2.0"

# TODO extend for other architectures
_SDK_ARCH = ["amd64"]


class DotNetBadArchitectureError(errors.SnapcraftError):

    fmt = (
        "Failed to prepare the .NET SDK: "
        "The architecture {architecture!r} is not supported. "
        "Supported architectures are: {supported}."
    )

    def __init__(self, *, architecture: str, supported: List[str]) -> None:
        super().__init__(
            architecture=architecture,
            supported=formatting_utils.humanize_list(supported, "and"),
        )


class DotNetBadReleaseDataError(errors.SnapcraftError):

    fmt = (
        "Failed to prepare the .NET SDK: "
        "An error occurred while fetching the version details "
        "for {version!r}. Check that the version is correct."
    )

    def __init__(self, version):
        super().__init__(version=version)


class DotNetPlugin(PluginV1):
    @classmethod
    def schema(cls):
        schema = super().schema()

        schema["properties"]["dotnet-version"] = {
            "type": "number",
            "default": _VERSION_DEFAULT,
        }
        schema["properties"]["dotnet-runtime-version"] = {
            "type": "string",
            "default": _RUNTIME_DEFAULT,
        }
        schema["required"] = ["source"]

        return schema

    @classmethod
    def get_pull_properties(cls):
        # Inform Snapcraft of the properties associated with pulling. If these
        # change in the YAML Snapcraft will consider the build step dirty.
        return ["dotnet-runtime-version", "dotnet-version"]

    def __init__(self, name, options, project):
        super().__init__(name, options, project)

        self._dotnet_dir = os.path.join(self.partdir, "dotnet")
        self._dotnet_sdk_dir = os.path.join(self._dotnet_dir, "sdk")

        self._setup_base_tools(project._get_build_base())

        self._sdk = self._get_sdk()
        self._dotnet_cmd = os.path.join(self._dotnet_sdk_dir, "dotnet")

    def _setup_base_tools(self, base):
        extra_packages = [
            "libcurl3",
            "libcurl3-gnutls",
            "liblttng-ust0",
            "libunwind8",
            "lldb",
            "libssl1.0.0",
            "libgssapi-krb5-2",
            "zlib1g",
            "libgcc1",
        ]
        if base in ("core", "core16"):
            self.stage_packages += extra_packages + ["libicu55"]
        elif base in ("core18",):
            self.stage_packages += extra_packages + ["libicu60"]
        else:
            raise errors.PluginBaseError(part_name=self.name, base=base)

    def _get_sdk(self):
        sdk_arch = self.project.deb_arch
        if sdk_arch not in _SDK_ARCH:
            raise DotNetBadArchitectureError(architecture=sdk_arch, supported=_SDK_ARCH)
        # TODO: Make this a class that takes care of retrieving the infos
        sdk_files = self._get_sdk_info(self.options.dotnet_runtime_version)

        return sources.Tar(sdk_files["url"], self._dotnet_sdk_dir)

    def pull(self):
        super().pull()

        os.makedirs(self._dotnet_sdk_dir, exist_ok=True)

        self._sdk.pull()

    def clean_pull(self):
        super().clean_pull()

        # Remove the dotnet directory (if any)
        if os.path.exists(self._dotnet_dir):
            shutil.rmtree(self._dotnet_dir)

    def build(self):
        super().build()

        if "debug" in self.options.build_attributes:
            configuration = "Debug"
        else:
            configuration = "Release"

        self.run([self._dotnet_cmd, "build", "-c", configuration])

        publish_cmd = [
            self._dotnet_cmd,
            "publish",
            "-c",
            configuration,
            "-o",
            self.installdir,
        ]
        # Build command for self-contained application
        publish_cmd += ["--self-contained", "-r", "linux-x64"]
        self.run(publish_cmd)

        # Workaround to set the right permission for the executable.
        appname = os.path.join(self.installdir, self._get_appname())
        if os.path.exists(appname):
            os.chmod(appname, 0o755)

    def _get_appname(self):
        for file in os.listdir(self.builddir):
            if fnmatch.fnmatch(file, "*.??proj"):
                return os.path.splitext(file)[0]

    def _get_version_metadata(self, version):
        json_data = self._get_dotnet_release_metadata()
        package_data = list(
            filter(lambda x: x.get("release-version") == version, json_data["releases"])
        )

        if not package_data:
            raise DotNetBadReleaseDataError(version)

        return package_data[0]

    def _get_dotnet_release_metadata(self):
        package_metadata = []

        metadata_url = _DOTNET_RELEASE_METADATA_URL.format(
            version=self.options.dotnet_version
        )
        req = urllib.request.Request(metadata_url)
        r = urllib.request.urlopen(req).read()
        package_metadata = json.loads(r.decode("utf-8"))

        return package_metadata

    def _get_sdk_info(self, version):
        metadata = self._get_version_metadata(version)

        sdk_files = list(
            filter(lambda x: x.get("rid") == "linux-x64", metadata["sdk"]["files"])
        )

        if not sdk_files:
            raise DotNetBadReleaseDataError(version)

        return sdk_files[0]

    def env(self, root):
        # Update the PATH only during the Build and Install step
        if root == self.installdir:
            return ["PATH={}:$PATH".format(self._dotnet_sdk_dir)]
        else:
            return []



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Create snaps from conda packages.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - conda-packages
      (list of strings, default [])
      List of conda packages to install.

    - conda-python-version
      string
      Python version major and minor version (e.g. 3.8).

    - conda-miniconda-version
      string, default latest
      The version of miniconda to initialize.
"""

import os
import platform
from textwrap import dedent
from typing import Any, Dict, List, Set

from snapcraft.internal.errors import SnapcraftException
from snapcraft.plugins.v2 import PluginV2


_MINICONDA_ARCH_FROM_SNAP_ARCH = {
    "i386": "x86",
    "amd64": "x86_64",
    "armhf": "armv7l",
    "ppc64el": "ppc64le",
}
_MINICONDA_ARCH_FROM_PLATFORM = {"x86_64": {"32bit": "x86", "64bit": "x86_64"}}


class ArchitectureMissing(SnapcraftException):
    def __init__(self, snap_arch: str) -> None:
        self.snap_arch = snap_arch

    def get_brief(self) -> str:
        return (
            f"Architecture {self.snap_arch!r} is not supported with the 'conda' plugin."
        )

    def get_resolution(self) -> str:
        return "Ensure running the build on a supported architecture for this plugin."


def _get_architecture() -> str:
    snap_arch = os.getenv("SNAP_ARCH")
    # The first scenario is the general case as snapcraft will be running from the snap.
    if snap_arch is not None:
        try:
            miniconda_arch = _MINICONDA_ARCH_FROM_SNAP_ARCH[snap_arch]
        except KeyError:
            raise ArchitectureMissing(snap_arch)
    # But there may be times when running from a virtualenv while doing development.
    else:
        miniconda_arch = _MINICONDA_ARCH_FROM_PLATFORM[platform.machine()][
            platform.architecture()[0]
        ]

    return miniconda_arch


def _get_miniconda_source(version: str) -> str:
    """Return tuple of source_url and source_checksum (if known)."""
    arch = _get_architecture()
    source = f"https://repo.anaconda.com/miniconda/Miniconda3-{version}-Linux-{arch}.sh"
    return source


class CondaPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "conda-packages": {
                    "type": "array",
                    "minItems": 1,
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "conda-python-version": {"type": "string", "default": ""},
                "conda-miniconda-version": {"type": "string", "default": "latest"},
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"curl"}

    def get_build_environment(self) -> Dict[str, str]:
        return {"PATH": "${HOME}/miniconda/bin:${PATH}"}

    def _get_download_miniconda_command(self, url: str) -> str:
        return dedent(
            f"""\
        if ! [ -e "${{HOME}}/miniconda.sh" ]; then
            curl --proto '=https' --tlsv1.2 -sSf {url} > ${{HOME}}/miniconda.sh
            chmod 755 ${{HOME}}/miniconda.sh
            export PATH="${{HOME}}/miniconda/bin:${{PATH}}"
        fi
        """
        )

    def _get_install_env_command(self) -> str:
        cmd = ["${HOME}/miniconda.sh", "-bfp", "${HOME}/miniconda"]
        return " ".join(cmd)

    def _get_deploy_command(self) -> str:
        conda_target_prefix = "/snap/${SNAPCRAFT_PROJECT_NAME}/current"

        deploy_cmd = [
            "CONDA_TARGET_PREFIX_OVERRIDE=" + conda_target_prefix,
            "conda",
            "create",
            "--prefix",
            "$SNAPCRAFT_PART_INSTALL",
            "--yes",
        ]
        if self.options.conda_python_version:
            deploy_cmd.append("python={}".format(self.options.conda_python_version))

        deploy_cmd.extend(self.options.conda_packages)

        return " ".join(deploy_cmd)

    def get_build_commands(self) -> List[str]:
        url = _get_miniconda_source(self.options.conda_miniconda_version)
        return [
            self._get_download_miniconda_command(url),
            self._get_install_env_command(),
            self._get_deploy_command(),
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import fixtures
from testtools.matchers import Equals

from snapcraft.internal import repo
from snapcraft.internal.pluginhandler._dependencies import MissingDependencyResolver
from tests import unit


class MissingDependencyTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        packages = {"/bin/bash": "bash", "/bin/sh": "dash"}

        def fake_repo_query(*args, **kwargs):
            file_path = kwargs["file_path"]
            try:
                return packages[file_path]
            except KeyError:
                raise repo.errors.FileProviderNotFound(file_path=file_path)

        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.internal.repo.Repo.get_package_for_file",
                side_effect=fake_repo_query,
            )
        )

    def test_all_stage_packages(self):
        d = MissingDependencyResolver(elf_files=["/bin/bash", "/bin/sh"])

        self.assertThat(d._stage_packages_dependencies, Equals({"bash", "dash"}))
        self.assertThat(d._unhandled_dependencies, Equals(set()))

        echoer = mock.Mock()
        d.print_resolutions(
            part_name="fake-part", stage_packages_exist=True, echoer=echoer
        )
        echoer.warning.assert_called_once_with(
            "The 'fake-part' part is missing libraries that are not included "
            "in the snap or base. They can be satisfied by adding the "
            "following entries to the existing stage-packages for this part:\n- bash\n- dash"
        )

        echoer.reset_mock()
        d.print_resolutions(
            part_name="fake-part", stage_packages_exist=False, echoer=echoer
        )
        echoer.warning.assert_called_once_with(
            "The 'fake-part' part is missing libraries that are not included "
            "in the snap or base. They can be satisfied by adding the "
            "following entry for this part\nstage-packages:\n- bash\n- dash"
        )

    def test_unhandled_dependencies(self):
        d = MissingDependencyResolver(elf_files=["/foo", "/bar", "/foo/bar"])

        self.assertThat(d._stage_packages_dependencies, Equals(set()))
        self.assertThat(d._unhandled_dependencies, Equals({"/foo", "/bar", "/foo/bar"}))

        echoer = mock.Mock()
        d.print_resolutions(
            part_name="fake-part", stage_packages_exist=True, echoer=echoer
        )
        echoer.warning.assert_called_once_with(
            "This part is missing libraries that cannot be satisfied with "
            "any available stage-packages known to snapcraft:\n"
            "- /bar\n- /foo\n- /foo/bar\n"
            "These dependencies can be satisfied via additional parts or "
            "content sharing. "
            "Consider validating configured filesets if this dependency was built."
        )

    def test_stage_packages_and_unhandled_dependencies(self):
        d = MissingDependencyResolver(
            elf_files=["/bin/bash", "/bin/sh", "/foo", "/bar", "/foo/bar"]
        )

        self.assertThat(d._stage_packages_dependencies, Equals({"bash", "dash"}))
        self.assertThat(d._unhandled_dependencies, Equals({"/foo", "/bar", "/foo/bar"}))

        echoer = mock.Mock()
        d.print_resolutions(
            part_name="fake-part", stage_packages_exist=False, echoer=echoer
        )
        echoer.warning.assert_has_calls(
            [
                mock.call(
                    "The 'fake-part' part is missing libraries that are not included "
                    "in the snap or base. They can be satisfied by adding the "
                    "following entry for this part\nstage-packages:\n- bash\n- dash"
                ),
                mock.call(
                    "This part is missing libraries that cannot be satisfied with "
                    "any available stage-packages known to snapcraft:\n"
                    "- /bar\n- /foo\n- /foo/bar\n"
                    "These dependencies can be satisfied via additional parts or "
                    "content sharing. "
                    "Consider validating configured filesets if this dependency was built."
                ),
            ]
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

import testscenarios
from testtools.matchers import Contains, Equals

from tests import fixture_setup, integration
from tests.integration import repo


class BuildSnapsTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = (
        ("snap name", {"snap": "u1test-snap-with-tracks"}),
        (
            "snap name with track and risk",
            {"snap": "u1test-snap-with-tracks/test-track-1/beta"},
        ),
    )

    def test_build_snap(self):
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("The autopkgtest armhf runners can't install snaps")
        self.useFixture(fixture_setup.WithoutSnapInstalled(self.snap))
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "test-part-with-build-snap", {"plugin": "nil", "build-snaps": [self.snap]}
        )
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft("build")
        self.assertTrue(repo.is_snap_installed(self.snap))


class BuildSnapsErrorsTestCase(integration.TestCase):
    def test_inexistent_build_snap(self):
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("The autopkgtest armhf runners can't install snaps")
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "test-part-with-build-snap",
            {"plugin": "nil", "build-snaps": ["inexistent"]},
        )
        self.useFixture(snapcraft_yaml)

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, ["build"]
        )

        self.assertThat(exception.returncode, Equals(2))
        self.assertThat(exception.output, Contains("'inexistent'"))
        self.assertFalse(repo.is_snap_installed("inexistent"))

    def test_snap_exists_but_not_on_channel(self):
        # If the snap tested here does not exist, then BuildSnapsTestCase
        # will fail.
        if os.environ.get("ADT_TEST") and self.deb_arch == "armhf":
            self.skipTest("The autopkgtest armhf runners can't install snaps")
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part(
            "test-part-with-build-snap",
            {
                "plugin": "nil",
                "build-snaps": ["u1test-snap-with-tracks/not-exists/candidate"],
            },
        )
        self.useFixture(snapcraft_yaml)

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, ["build"]
        )

        self.assertThat(exception.returncode, Equals(2))
        self.assertThat(exception.output, Contains("'u1test-snap-with-tracks'"))

        self.assertFalse(repo.is_snap_installed("u1test-snap-with-tracks"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

import fixtures
import testscenarios
from testtools.matchers import FileExists, Not

from tests import fixture_setup, integration


class ParserTestCase(integration.TestCase):
    """Test bin/snapcraft-parser"""

    def setUp(self):
        super().setUp()
        if os.getenv("SNAPCRAFT_PACKAGE_TYPE") == "snap":
            self.skipTest("The snapcraft-parser is not provided by the snap")
        self.useFixture(fixtures.EnvironmentVariable("TMPDIR", self.path))

    def call_parser(self, wiki_path, expect_valid, expect_output=True):
        part_file = os.path.join(self.path, "parts.yaml")
        args = ["--index", wiki_path, "--output", part_file]

        if expect_valid:
            self.run_snapcraft_parser(args)
        else:
            self.assertRaises(
                subprocess.CalledProcessError, self.run_snapcraft_parser, args
            )

        if expect_output:
            self.assertThat(part_file, FileExists())
        else:
            self.assertThat(part_file, Not(FileExists()))


class TestParser(ParserTestCase):
    """Test bin/snapcraft-parser"""

    def test_parser_basic(self):
        """Test snapcraft-parser basic usage"""
        fixture = fixture_setup.FakePartsWiki()
        self.useFixture(fixture)

        self.call_parser(fixture.fake_parts_wiki_fixture.url, expect_valid=True)


class TestParserWikis(testscenarios.WithScenarios, ParserTestCase):
    """Test bin/snapcraft-parser"""

    def setUp(self):
        temp_cwd_fixture = fixture_setup.TempCWD()
        self.useFixture(temp_cwd_fixture)
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable(
                "XDG_CACHE_HOME", os.path.join(self.path, ".cache")
            )
        )

    def _read_file(self, filename):
        content = ""
        with open(filename) as fp:
            content = fp.read()

        return content

    def _setup_wiki_file(self, filename, origin, commit=None):
        content = self._read_file(filename)
        content = content.replace("$ORIGIN", origin)
        if commit:
            content = content.replace("$COMMIT", commit)
        with open("wiki", "w") as fp:
            fp.write(content)

    def _setup_origin(self, snapcraft_files, repo_dir, base_dir):
        os.makedirs(repo_dir)
        previous_dir = os.getcwd()
        os.chdir(repo_dir)

        subprocess.check_call(
            ["git", "init", "."], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        subprocess.check_output(["git", "config", "user.name", "Test User"])
        subprocess.check_output(
            ["git", "config", "user.email", "<test.user@example.com"]
        )

        for snapcraft_file in snapcraft_files:
            snapcraft_content_path = os.path.join(base_dir, snapcraft_file["path"])
            snapcraft_content = self._read_file(snapcraft_content_path)
            snapcraft_file = snapcraft_file["snapcraft_file"]
            # make subdirectories if needed.
            if os.path.dirname(snapcraft_file):
                os.makedirs(os.path.dirname(snapcraft_file))
            with open(snapcraft_file, "w") as fp:
                fp.write(snapcraft_content)
            subprocess.check_call(
                ["git", "add", snapcraft_file],
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
            )
        subprocess.check_call(
            ["git", "commit", "-am", "commit"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        subprocess.check_call(["git", "branch", "feature-branch"])
        commit = (
            subprocess.check_output(["git", "log"])
            .split()[0]
            .decode("utf-8")
            .replace("commit: ", "")
        )

        os.chdir(previous_dir)
        return commit

    scenarios = [
        (
            "Hidden .snapcraft.yaml file",
            {
                "wiki_file": "hidden_parts_wiki",
                "expect_valid": True,
                "expect_output": True,
                "snapcraft_files": [
                    {
                        "path": "hidden_snapcraft_yaml",
                        "snapcraft_file": ".snapcraft.yaml",
                    }
                ],
            },
        ),
        (
            "Both snapcraft.yaml and .snapcraft.yaml files",
            {
                "wiki_file": "both_parts_wiki",
                "expect_valid": False,
                "expect_output": True,
                "snapcraft_files": [
                    {
                        "path": "hidden_snapcraft_yaml",
                        "snapcraft_file": ".snapcraft.yaml",
                    },
                    {
                        "path": "hidden_snapcraft_yaml",
                        "snapcraft_file": "snapcraft.yaml",
                    },
                ],
            },
        ),
        (
            "snap/snapcraft.yaml",
            {
                "wiki_file": "snap.snapcraft.yaml",
                "expect_valid": True,
                "expect_output": True,
                "snapcraft_files": [
                    {
                        "path": "hidden_snapcraft_yaml",
                        "snapcraft_file": "snap/snapcraft.yaml",
                    }
                ],
            },
        ),
        (
            "Missing .snapcraft.yaml file",
            {
                "wiki_file": "missing_parts_wiki",
                "expect_valid": False,
                "expect_output": True,
                "snapcraft_files": [
                    {
                        "path": "hidden_snapcraft_yaml",
                        "snapcraft_file": "missing_snapcraft.yaml",
                    }
                ],
            },
        ),
        (
            "Origin type, branch and commit options",
            {
                "wiki_file": "origin_options_wiki",
                "expect_valid": True,
                "expect_output": True,
                "snapcraft_files": [
                    {
                        "path": "origin_options_snapcraft_yaml",
                        "snapcraft_file": "snapcraft.yaml",
                    }
                ],
            },
        ),
        (
            "Origin type, branch and commit options (wrong values)",
            {
                "wiki_file": "wrong_origin_options_wiki",
                "expect_valid": False,
                "expect_output": False,
                "snapcraft_files": [
                    {
                        "path": "wrong_origin_options_snapcraft_yaml",
                        "snapcraft_file": "snapcraft.yaml",
                    }
                ],
            },
        ),
    ]

    def test_local_wiki(self):
        repo_dir = "repo"
        base_dir = os.path.join(os.path.dirname(integration.__file__), "wiki")
        wiki_file = os.path.join(base_dir, self.wiki_file)

        commit = self._setup_origin(self.snapcraft_files, repo_dir, base_dir)
        self._setup_wiki_file(wiki_file, repo_dir, commit)
        self.call_parser("wiki", self.expect_valid, self.expect_output)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from testtools.matchers import Equals, FileExists

from tests import integration


class HgSourceTestCase(integration.HgSourceBaseTestCase):
    def test_pull_hg_head(self):
        self.copy_project_to_cwd("hg-head")

        self.init_source_control()
        open("1", "w").close()
        self.commit("1", "1")
        open("2", "w").close()
        self.commit("2", "2")

        self.run_snapcraft("pull")
        revno = self.get_revno(os.path.join(self.parts_dir, "mercurial", "src"))
        self.assertThat(revno, Equals('"2"'))

        self.run_snapcraft("pull")
        revno = self.get_revno(os.path.join(self.parts_dir, "mercurial", "src"))
        self.assertThat(revno, Equals('"2"'))

    def test_pull_hg_tag(self):
        self.copy_project_to_cwd("hg-tag")

        self.init_source_control()
        open("1", "w").close()
        self.commit("1", "1")
        subprocess.check_call(["hg", "tag", "initial", "--user", '"Example Dev"'])
        open("2", "w").close()
        self.commit("2", "2")

        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            "ls -1 {} | wc -l ".format(
                os.path.join(self.parts_dir, "mercurial", "src")
            ),
            shell=True,
            universal_newlines=True,
        ).strip()
        self.assertThat(revno, Equals("1"))

        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            "ls -1 {} | wc -l ".format(
                os.path.join(self.parts_dir, "mercurial", "src")
            ),
            shell=True,
            universal_newlines=True,
        ).strip()
        self.assertThat(revno, Equals("1"))

    def test_pull_hg_commit(self):
        self.copy_project_to_cwd("hg-commit")

        self.init_source_control()
        open("1", "w").close()
        self.commit("1", "1")
        open("2", "w").close()
        self.commit("2", "2")

        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            "ls -1 {} | wc -l ".format(
                os.path.join(self.parts_dir, "mercurial", "src")
            ),
            shell=True,
            universal_newlines=True,
        ).strip()
        self.assertThat(revno, Equals("1"))

        self.run_snapcraft("pull")
        revno = subprocess.check_output(
            "ls -1 {} | wc -l ".format(
                os.path.join(self.parts_dir, "mercurial", "src")
            ),
            shell=True,
            universal_newlines=True,
        ).strip()
        self.assertThat(revno, Equals("1"))

    def test_pull_hg_branch(self):
        self.copy_project_to_cwd("hg-branch")

        self.init_source_control()
        subprocess.check_call(["hg", "branch", "second"], stdout=subprocess.DEVNULL)
        open("second", "w").close()
        self.commit("second", "second")
        subprocess.check_call(["hg", "branch", "default"], stdout=subprocess.DEVNULL)
        open("default", "w").close()
        self.commit("default", "default")

        self.run_snapcraft("pull")
        self.assertThat(
            os.path.join(self.parts_dir, "mercurial", "src", "second"), FileExists()
        )

        self.run_snapcraft("pull")
        self.assertThat(
            os.path.join(self.parts_dir, "mercurial", "src", "second"), FileExists()
        )
